"""This file contains unit tests of XarrayDataset."""

import dataclasses
import datetime
from collections import namedtuple
from collections.abc import Iterable

import cftime
import numpy as np
import pandas as pd
import pytest
import torch
import xarray as xr
from xarray.coding.times import CFDatetimeCoder

from fme.core.coordinates import (
    DepthCoordinate,
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
    NullVerticalCoordinate,
)
from fme.core.dataset.concat import XarrayConcat
from fme.core.dataset.config import (
    FillNaNsConfig,
    OverwriteConfig,
    RepeatedInterval,
    TimeSlice,
    XarrayDataConfig,
)
from fme.core.dataset.getters import get_dataset, get_xarray_dataset
from fme.core.dataset.merged import MergedXarrayDataset
from fme.core.dataset.subset import XarraySubset
from fme.core.dataset.xarray import (
    XarrayDataset,
    _get_cumulative_timesteps,
    _get_file_local_index,
    _get_raw_times,
    _get_timestep,
    _get_vertical_coordinate,
    _repeat_and_increment_time,
)
from fme.core.typing_ import Slice

from .utils import as_broadcasted_tensor

SLICE_NONE = slice(None)
MOCK_DATA_FREQ = "3h"
MOCK_DATA_START_DATE = "2003-03"
MOCK_DATA_LAT_DIM, MOCK_DATA_LON_DIM = ("lat", "lon")


@dataclasses.dataclass
class VariableNames:
    time_dependent_names: Iterable[str]
    time_invariant_names: Iterable[str]
    initial_condition_names: Iterable[str]

    def _concat(self, *lists):
        return_value = []
        for list in lists:
            return_value.extend(list)
        return return_value

    @property
    def all_names(self) -> list[str]:
        return self._concat(
            self.time_dependent_names,
            self.time_invariant_names,
            self.initial_condition_names,
        )

    @property
    def spatial_resolved_names(self) -> list[str]:
        return self._concat(self.time_dependent_names, self.time_invariant_names)


MockData = namedtuple(
    "MockData", ("tmpdir", "obs_times", "start_times", "start_indices", "var_names")
)


def _get_data(
    tmp_path_factory,
    dirname,
    start,
    end,
    file_freq,
    step_freq,
    calendar,
    with_nans=False,
    var_names=["foo", "bar"],
    write_extra_vars=True,
    add_ensemble_dim=False,
) -> MockData:
    """Constructs an xarray dataset and saves to disk in netcdf format."""
    obs_times = xr.date_range(
        start,
        end,
        freq=step_freq,
        calendar=calendar,
        inclusive="left",
        use_cftime=True,
    )
    start_times = xr.date_range(
        start,
        end,
        freq=file_freq,
        calendar=calendar,
        inclusive="left",
        use_cftime=True,
    )
    obs_delta = obs_times[1] - obs_times[0]
    n_levels = 2
    n_lat, n_lon = 4, 8
    n_sample = 3

    non_time_dims = ("sample", "lat", "lon") if add_ensemble_dim else ("lat", "lon")
    non_time_shape = (n_sample, n_lat, n_lon) if add_ensemble_dim else (n_lat, n_lon)

    constant_var = xr.DataArray(
        np.random.randn(*non_time_shape).astype(np.float32),
        dims=non_time_dims,
    )
    constant_scalar_var = xr.DataArray(1.0).astype(np.float32)
    ak = {f"ak_{i}": float(i) for i in range(n_levels)}
    bk = {f"bk_{i}": float(i + 1) for i in range(n_levels)}
    tmpdir = tmp_path_factory.mktemp(dirname)
    filenames = []
    for i, first in enumerate(start_times):
        if first != start_times[-1]:
            last = start_times[i + 1]
        else:
            last = obs_times[-1] + obs_delta
        time = xr.date_range(
            first,
            last,
            freq=step_freq,
            calendar=calendar,
            inclusive="left",
            use_cftime=True,
        )
        data_vars: dict[str, float | xr.DataArray] = {**ak, **bk}
        for var_name in var_names:
            data = np.random.randn(len(time), *non_time_shape).astype(np.float32)
            if with_nans:
                data[0, :, 0] = np.nan
            data_vars[var_name] = xr.DataArray(data, dims=("time", *non_time_dims))

        data_varying_scalar = np.random.randn(len(time)).astype(np.float32)
        if with_nans:
            constant_var[0, 0] = np.nan

        if write_extra_vars:
            data_vars["varying_scalar_var"] = xr.DataArray(
                data_varying_scalar, dims=("time",)
            )
            data_vars["constant_var"] = constant_var
            data_vars["constant_scalar_var"] = constant_scalar_var

        coords = {
            "time": xr.DataArray(time, dims=("time",)),
            "lat": xr.DataArray(np.arange(n_lat, dtype=np.float32), dims=("lat",)),
            "lon": xr.DataArray(np.arange(n_lon, dtype=np.float32), dims=("lon",)),
        }
        if add_ensemble_dim:
            coords["sample"] = xr.DataArray(
                np.arange(n_sample, dtype=np.float32), dims=("sample",)
            )
            # variable without the ensemble dimension is useful for checking
            # broadcast behavior
            data_vars["var_no_ensemble_dim"] = xr.DataArray(
                np.random.randn(len(time), n_lat, n_lon).astype(np.float32),
                dims=("time", "lat", "lon"),
            )
            # set values to sample index for testing convenience
            sample_index_values = np.broadcast_to(
                np.arange(n_sample).reshape(1, n_sample, 1, 1),  # shape [1, ns, 1, 1],
                (len(time), n_sample, n_lat, n_lon),
            )
            data_vars["var_matches_sample_index"] = (
                xr.zeros_like(data_vars["foo"]) + sample_index_values
            )

        ds = xr.Dataset(data_vars=data_vars, coords=coords)
        filename = tmpdir / f"{first.strftime('%Y%m%d%H')}.nc"
        ds.to_netcdf(
            filename,
            unlimited_dims=["time"],
            format="NETCDF4",
        )
        filenames.append(filename)

    initial_condition_names = ()
    start_indices = _get_cumulative_timesteps(_get_raw_times(filenames, "netcdf4"))
    if write_extra_vars:
        variable_names = VariableNames(
            time_dependent_names=(*var_names, "varying_scalar_var"),
            time_invariant_names=("constant_var", "constant_scalar_var"),
            initial_condition_names=initial_condition_names,
        )
    else:
        variable_names = VariableNames(
            time_dependent_names=var_names,
            time_invariant_names=(),
            initial_condition_names=initial_condition_names,
        )
    return MockData(tmpdir, obs_times, start_times, start_indices, variable_names)


def get_mock_monthly_netcdfs(
    tmp_path_factory,
    dirname,
    with_nans=False,
    end_date="2003-06",
    var_names=["foo", "bar"],
    write_extra_vars=True,
    add_ensemble_dim=False,
) -> MockData:
    return _get_data(
        tmp_path_factory,
        dirname,
        start=MOCK_DATA_START_DATE,
        end=end_date,
        file_freq="MS",
        step_freq=MOCK_DATA_FREQ,
        calendar="standard",
        with_nans=with_nans,
        var_names=var_names,
        write_extra_vars=write_extra_vars,
        add_ensemble_dim=add_ensemble_dim,
    )


@pytest.fixture(scope="session")
def mock_monthly_netcdfs(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(tmp_path_factory, "month")


@pytest.fixture(scope="session")
def mock_monthly_netcdfs_another_source(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(
        tmp_path_factory, "month_another_source", var_names=["baz", "qux"]
    )


@pytest.fixture(scope="session")
def mock_monthly_netcdfs_another_source_diff_time(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(
        tmp_path_factory,
        "month_another_source",
        end_date="2003-08",
        var_names=["baz", "qux"],
        write_extra_vars=False,
    )


@pytest.fixture(scope="session")
def mock_monthly_netcdfs_with_nans(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(tmp_path_factory, "month_with_nans", with_nans=True)


@pytest.fixture(scope="session")
def mock_monthly_netcdfs_ensemble_dim(tmp_path_factory) -> MockData:
    return get_mock_monthly_netcdfs(
        tmp_path_factory,
        "month_with_ensemble_dim",
        add_ensemble_dim=True,
        var_names=["foo", "bar", "var_no_ensemble_dim", "var_matches_sample_index"],
    )


@pytest.fixture(scope="session")
def mock_monthly_zarr_ensemble_dim(
    tmp_path_factory, mock_monthly_netcdfs_ensemble_dim
) -> MockData:
    zarr_parent = tmp_path_factory.mktemp("zarr")
    filename = "data.zarr"
    data = load_files_without_dask(
        mock_monthly_netcdfs_ensemble_dim.tmpdir.glob("*.nc")
    )
    data.to_zarr(zarr_parent / filename)
    return MockData(
        zarr_parent,
        mock_monthly_netcdfs_ensemble_dim.obs_times,
        mock_monthly_netcdfs_ensemble_dim.start_times,
        mock_monthly_netcdfs_ensemble_dim.start_indices,
        mock_monthly_netcdfs_ensemble_dim.var_names,
    )


def load_files_without_dask(files, engine="netcdf4") -> xr.Dataset:
    """Load a sequence of files without dask, concatenating along the time dimension.

    We load the data from the files into memory to ensure Datasets are properly closed,
    since xarray cannot concatenate Datasets lazily without dask anyway. This should be
    acceptable for the small datasets we construct for test purposes.
    """
    datasets = []
    for file in sorted(files):
        with xr.open_dataset(
            file,
            decode_times=CFDatetimeCoder(use_cftime=True),
            decode_timedelta=False,
            engine=engine,
        ) as ds:
            datasets.append(ds.load())
    return xr.concat(datasets, dim="time", data_vars="minimal", coords="minimal")


@pytest.fixture(scope="session")
def mock_monthly_zarr(tmp_path_factory, mock_monthly_netcdfs) -> MockData:
    zarr_parent = tmp_path_factory.mktemp("zarr")
    filename = "data.zarr"
    data = load_files_without_dask(mock_monthly_netcdfs.tmpdir.glob("*.nc"))
    data.to_zarr(zarr_parent / filename)
    return MockData(
        zarr_parent,
        mock_monthly_netcdfs.obs_times,
        mock_monthly_netcdfs.start_times,
        mock_monthly_netcdfs.start_indices,
        mock_monthly_netcdfs.var_names,
    )


@pytest.fixture(scope="session")
def mock_monthly_zarr_with_nans(
    tmp_path_factory, mock_monthly_netcdfs_with_nans
) -> MockData:
    zarr_parent = tmp_path_factory.mktemp("zarr")
    filename = "data.zarr"
    data = load_files_without_dask(mock_monthly_netcdfs_with_nans.tmpdir.glob("*.nc"))
    data.to_zarr(zarr_parent / filename)
    return MockData(
        zarr_parent,
        mock_monthly_netcdfs_with_nans.obs_times,
        mock_monthly_netcdfs_with_nans.start_times,
        mock_monthly_netcdfs_with_nans.start_indices,
        mock_monthly_netcdfs_with_nans.var_names,
    )


@pytest.fixture(scope="session")
def mock_yearly_netcdfs(tmp_path_factory):
    return _get_data(
        tmp_path_factory,
        "yearly",
        start="1999",
        end="2001",
        file_freq="YS",
        step_freq="1D",
        calendar="noleap",
    )


@pytest.mark.parametrize(
    "global_idx,expected_file_local_idx",
    [
        pytest.param(0, (0, 0), id="monthly_file_local_idx_2003_03_01_00"),
        pytest.param(1, (0, 1), id="monthly_file_local_idx_2003_03_01_03"),
        pytest.param(30 * 8, (0, 30 * 8), id="monthly_file_local_idx_2003_03_31_00"),
        pytest.param(31 * 8, (1, 0), id="monthly_file_local_idx_2003_04_01_00"),
        pytest.param(
            (31 + 30 + 20) * 8 - 1,
            (2, 20 * 8 - 1),
            id="monthly_file_local_idx_2003_05_20_21",
        ),
    ],
)
def test_monthly_file_local_index(
    mock_monthly_netcdfs, global_idx, expected_file_local_idx
):
    mock_data: MockData = mock_monthly_netcdfs
    file_local_idx = _get_file_local_index(global_idx, mock_data.start_indices)
    assert file_local_idx == expected_file_local_idx
    delta = mock_data.obs_times[1] - mock_data.obs_times[0]
    target_timestamp = np.datetime64(
        cftime.DatetimeGregorian(2003, 3, 1, 0, 0, 0, 0, has_year_zero=False)
        + global_idx * delta
    )
    file_idx, local_idx = file_local_idx
    full_paths = sorted(list(mock_data.tmpdir.glob("*.nc")))
    with xr.open_dataset(
        full_paths[file_idx],
        decode_times=CFDatetimeCoder(use_cftime=True),
        decode_timedelta=False,
    ) as ds:
        assert ds["time"][local_idx].item() == target_timestamp


def _test_monthly_values(
    mock_data: MockData,
    global_idx,
    expected_n_samples=None,
    file_pattern="*.nc",
    engine="netcdf4",
):
    """Runs shape and length checks on the dataset."""
    var_names: VariableNames = mock_data.var_names
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir, file_pattern=file_pattern, engine=engine
    )
    dataset = XarrayDataset(config, var_names.all_names, 2)
    if expected_n_samples is None:
        expected_n_samples = len(mock_data.obs_times) - 1

    assert len(dataset) == expected_n_samples
    arrays, time = dataset[global_idx]
    ds = load_files_without_dask(mock_data.tmpdir.glob(file_pattern), engine=engine)
    target_times = ds["time"][global_idx : global_idx + 2].drop_vars("time")
    xr.testing.assert_equal(time, target_times)
    lat_dim, lon_dim = MOCK_DATA_LAT_DIM, MOCK_DATA_LON_DIM
    dims = ("time", str(lat_dim), str(lon_dim))
    shape = (2, ds.sizes[lat_dim], ds.sizes[lon_dim])
    time_slice = slice(global_idx, global_idx + 2)
    for var_name in var_names.spatial_resolved_names:
        data = arrays[var_name]
        assert data.shape[0] == 2
        da = ds[var_name]
        if var_name in var_names.time_dependent_names:
            da = da.isel(time=time_slice)
        target_data = as_broadcasted_tensor(da.variable, dims, shape)
        assert torch.equal(data, target_data)

    for var_name in mock_data.var_names.initial_condition_names:
        data = arrays[var_name].detach().numpy()
        assert data.shape[0] == 1
        target_data = ds[var_name][global_idx : global_idx + 1, :, :].values
        assert np.all(data == target_data)


@pytest.mark.parametrize(
    "global_idx",
    [
        pytest.param(31 * 8 - 1, id="monthly_XarrayDataset_2003_03_31_21"),
        pytest.param((31 + 30 + 20) * 8 - 1, id="monthly_XarrayDataset_2003_05_20_21"),
        pytest.param((31 + 30) * 8 - 1, id="2003_04_30_21 (test for GH #1942)"),
    ],
)
@pytest.mark.parametrize(
    "mock_data_fixture, engine, file_pattern",
    [
        ("mock_monthly_netcdfs", "netcdf4", "*.nc"),
        ("mock_monthly_zarr", "zarr", "*.zarr"),
    ],
)
def test_XarrayDataset_monthly(
    global_idx, mock_data_fixture, engine, file_pattern, request
):
    mock_data: MockData = request.getfixturevalue(mock_data_fixture)
    _test_monthly_values(
        mock_data, global_idx, file_pattern=file_pattern, engine=engine
    )


@pytest.mark.parametrize("n_samples", [None, 1])
def test_XarrayDataset_monthly_n_timesteps(mock_monthly_netcdfs, n_samples):
    """Test that increasing n_timesteps decreases the number of samples."""
    mock_data: MockData = mock_monthly_netcdfs
    if len(mock_data.var_names.initial_condition_names) != 0:
        return
    config = XarrayDataConfig(data_path=mock_data.tmpdir, subset=Slice(stop=n_samples))
    n_forward_steps = 4
    dataset, _ = get_xarray_dataset(
        config, mock_data.var_names.all_names + ["x"], n_forward_steps + 1
    )
    if n_samples is None:
        assert len(dataset) == len(mock_data.obs_times) - n_forward_steps
    else:
        assert len(dataset) == n_samples
    assert "x" in dataset[0][0]


@pytest.mark.parametrize(
    "global_idx,expected_file_local_idx",
    [
        pytest.param(365 + 59, (1, 59), id="yearly_file_local_idx_2000_03_01"),
        pytest.param(365, (1, 0), id="yearly_file_local_idx_2000_01_01"),
        pytest.param(364, (0, 364), id="yearly_file_local_idx_1999_12_31"),
    ],
)
def test_yearly_file_local_index(
    mock_yearly_netcdfs, global_idx, expected_file_local_idx
):
    mock_data: MockData = mock_yearly_netcdfs
    file_local_idx = _get_file_local_index(global_idx, mock_data.start_indices)
    assert file_local_idx == expected_file_local_idx
    delta = mock_data.obs_times[1] - mock_data.obs_times[0]
    target_timestamp = (
        cftime.DatetimeNoLeap(1999, 1, 1, 0, 0, 0, 0, has_year_zero=True)
        + global_idx * delta
    )
    file_idx, local_idx = file_local_idx
    full_paths = sorted(list(mock_data.tmpdir.glob("*.nc")))
    with xr.open_dataset(
        full_paths[file_idx],
        decode_times=CFDatetimeCoder(use_cftime=True),
        decode_timedelta=False,
    ) as ds:
        assert ds["time"][local_idx].item() == target_timestamp


@pytest.mark.parametrize(
    "global_idx",
    [
        pytest.param(364, id="yearly_XarrayDataset_1999_12_31"),
        pytest.param(365 + 31 + 28, id="yearly_XarrayDataset_2000_02_28"),
    ],
)
def test_XarrayDataset_yearly(mock_yearly_netcdfs, global_idx):
    mock_data: MockData = mock_yearly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir)
    ds = load_files_without_dask(mock_data.tmpdir.glob("*.nc"))
    for n_steps in [3, 50]:
        dataset = XarrayDataset(config, mock_data.var_names.all_names, n_steps)
        assert len(dataset) == len(mock_data.obs_times) - n_steps + 1
        lon_dim, lat_dim = MOCK_DATA_LON_DIM, MOCK_DATA_LAT_DIM
        dims = ("time", lat_dim, lon_dim)
        shape = (n_steps, ds.sizes[lat_dim], ds.sizes[lon_dim])
        time_slice = slice(global_idx, global_idx + n_steps)
        for var_name in mock_data.var_names.spatial_resolved_names:
            da = ds[var_name]
            if var_name in mock_data.var_names.time_dependent_names:
                da = da.isel(time=time_slice)
            target_data = as_broadcasted_tensor(da.variable, dims, shape)
            target_times = ds["time"][global_idx : global_idx + n_steps].drop_vars(
                "time"
            )
            data, time = dataset[global_idx]
            data_tensor = data[var_name]
            assert data_tensor.shape[0] == n_steps
            assert torch.equal(data_tensor, target_data)
            xr.testing.assert_equal(time, target_times)


def test_dataset_dtype_casting(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir, dtype="bfloat16")
    dataset = XarrayDataset(config, mock_data.var_names.all_names, 2)
    data_properties = dataset.properties
    assert isinstance(data_properties.horizontal_coordinates, LatLonCoordinates)
    assert data_properties.horizontal_coordinates.lat.dtype == torch.bfloat16
    assert data_properties.horizontal_coordinates.lon.dtype == torch.bfloat16
    assert isinstance(
        data_properties.vertical_coordinate, HybridSigmaPressureCoordinate
    )
    assert data_properties.vertical_coordinate.ak.dtype == torch.bfloat16
    assert data_properties.vertical_coordinate.bk.dtype == torch.bfloat16
    data, _ = dataset[0]
    for tensor in data.values():
        assert tensor.dtype == torch.bfloat16


def test_time_invariant_variable_is_repeated(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir)
    dataset = XarrayDataset(config, mock_data.var_names.all_names, 15)
    data = dataset[0][0]
    assert data["constant_var"].shape[0] == 15
    assert data["constant_scalar_var"].shape == (15, 4, 8)


def _get_repeat_dataset(
    mock_data: MockData, n_timesteps: int, n_repeats: int
) -> XarrayDataset:
    config = XarrayDataConfig(data_path=mock_data.tmpdir, n_repeats=n_repeats)
    return XarrayDataset(config, mock_data.var_names.all_names, n_timesteps)


@pytest.mark.parametrize("n_timesteps", [1, 4])
@pytest.mark.parametrize("n_repeats", [1, 2])
def test_repeat_dataset_num_timesteps(
    mock_monthly_netcdfs: MockData, n_timesteps, n_repeats
):
    unrepeated_dataset = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, 1)
    data = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, n_repeats)
    offset = n_timesteps - 1
    expected_length = n_repeats * (len(unrepeated_dataset) + offset) - offset
    assert len(data) == expected_length


@pytest.mark.parametrize(
    "glob_pattern, expected_num_files, expected_year_month_tuples",
    [
        ("*.nc", None, None),
        ("2003030100.nc", 1, [(2003, 3)]),
        ("2003??0100.nc", 3, [(2003, i) for i in range(3, 6)]),
    ],
    ids=["all_files", "single_file", "all_2003_files"],
)
def test_glob_file_pattern(
    mock_monthly_netcdfs: MockData,
    glob_pattern,
    expected_num_files,
    expected_year_month_tuples,
):
    config = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, file_pattern=glob_pattern
    )
    dataset = XarrayDataset(config, mock_monthly_netcdfs.var_names.all_names, 2)
    if expected_num_files is None:
        expected_num_files = len(mock_monthly_netcdfs.start_times)
    assert expected_num_files == len(dataset.full_paths)

    if expected_year_month_tuples is not None:
        for i, (year, month) in enumerate(expected_year_month_tuples):
            assert f"{year}{month:02d}" in dataset.full_paths[i]


def test_time_slice():
    time_slice = TimeSlice("2001-01-01", "2001-01-05", 2)
    time_index = xr.date_range(
        "2000", "2002", freq="D", calendar="noleap", use_cftime=True
    )
    slice_ = time_slice.slice(time_index)
    assert slice_ == slice(365, 370, 2)


def test_time_index(mock_monthly_netcdfs):
    config = XarrayDataConfig(data_path=mock_monthly_netcdfs.tmpdir)
    n_timesteps = 2
    names = mock_monthly_netcdfs.var_names.all_names
    dataset = XarrayDataset(config, names, n_timesteps)
    last_sample_init_time = len(mock_monthly_netcdfs.obs_times) - n_timesteps + 1
    obs_times = mock_monthly_netcdfs.obs_times[:last_sample_init_time]
    assert dataset.sample_start_times.equals(xr.CFTimeIndex(obs_times))


@pytest.mark.parametrize("infer_timestep", [True, False])
def test_XarrayDataset_timestep(mock_monthly_netcdfs, infer_timestep):
    config = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, infer_timestep=infer_timestep
    )
    names = mock_monthly_netcdfs.var_names.all_names
    n_timesteps = 2
    dataset = XarrayDataset(config, names, n_timesteps)
    if infer_timestep:
        expected_timestep = pd.Timedelta(MOCK_DATA_FREQ).to_pytimedelta()
        assert dataset.timestep == expected_timestep
    else:
        with pytest.raises(ValueError, match="Timestep was not inferred"):
            assert dataset.timestep


@pytest.mark.parametrize(
    ("periods", "freq", "reverse", "expected", "exception"),
    [
        pytest.param(
            2,
            "3h",
            False,
            datetime.timedelta(hours=3),
            None,
            id="2 timesteps, regular freq",
        ),
        pytest.param(
            3,
            "9h",
            False,
            datetime.timedelta(hours=9),
            None,
            id="3 timesteps, regular freq",
        ),
        pytest.param(3, "3h", True, None, ValueError, id="3 timesteps, negative freq"),
        pytest.param(
            3, "MS", False, None, ValueError, id="3 timesteps, irregular freq"
        ),
        pytest.param(1, "D", False, None, ValueError, id="1 timestep"),
    ],
)
def test_get_timestep(periods, freq, reverse, expected, exception):
    index = xr.date_range("2000", periods=periods, freq=freq, use_cftime=True)

    if reverse:
        index = index[::-1]

    if exception is None:
        result = _get_timestep(index.values)
        assert result == expected
    else:
        with pytest.raises(exception):
            _get_timestep(index.values)


@pytest.mark.parametrize("n_repeats", [1, 3])
def test_repeat_and_increment_times(n_repeats):
    freq = "5h"
    delta = pd.Timedelta(freq).to_pytimedelta()

    start_a = cftime.DatetimeGregorian(2000, 1, 1)
    periods_a = 2
    segment_a = xr.date_range(
        start_a, periods=periods_a, freq=freq, use_cftime=True
    ).values

    start_b = segment_a[-1] + delta
    periods_b = 3
    segment_b = xr.date_range(
        start_b, periods=periods_b, freq=freq, use_cftime=True
    ).values

    raw_times = [segment_a, segment_b]
    raw_periods = [periods_a, periods_b]
    raw_total_periods = sum(raw_periods)

    result = _repeat_and_increment_time(raw_times, n_repeats, delta)
    full_periods = [len(times) for times in result]
    full_total_periods = sum(full_periods)

    result_concatenated = np.concatenate(result)
    expected_concatenated = xr.date_range(
        start_a, periods=full_total_periods, freq=freq, use_cftime=True
    ).values

    assert full_periods == n_repeats * raw_periods
    assert full_total_periods == n_repeats * raw_total_periods
    np.testing.assert_equal(result_concatenated, expected_concatenated)


@pytest.mark.parametrize("n_repeats", [1, 3])
def test_all_times(mock_monthly_netcdfs, n_repeats):
    n_timesteps = 2  # Arbitrary for this test
    dataset = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, n_repeats)
    expected_periods = n_repeats * len(mock_monthly_netcdfs.obs_times)
    expected = xr.date_range(
        MOCK_DATA_START_DATE,
        periods=expected_periods,
        freq=MOCK_DATA_FREQ,
        use_cftime=True,
    )
    result = dataset.all_times
    assert result.equals(expected)


def test_get_sample_by_time_slice_times_n_repeats(mock_monthly_netcdfs: MockData):
    n_timesteps = 2  # Arbitrary for this test
    n_repeats = 3
    repeated_dataset = _get_repeat_dataset(mock_monthly_netcdfs, n_timesteps, n_repeats)

    # Pick a slice that is outside the range of the unrepeated data
    unrepeated_length = len(repeated_dataset.all_times) // n_repeats
    time_slice = slice(unrepeated_length, unrepeated_length + 3)

    _, result = repeated_dataset.get_sample_by_time_slice(time_slice)
    expected = xr.DataArray(
        repeated_dataset.all_times[time_slice].values, dims=["time"]
    )
    xr.testing.assert_equal(result, expected)


@pytest.mark.parametrize(
    "dtype,expected_torch_dtype", [("int16", torch.int16), (None, None)]
)
def test_dataset_config_dtype(dtype, expected_torch_dtype):
    config = XarrayDataConfig(data_path="path/to/data", dtype=dtype)
    assert config.torch_dtype == expected_torch_dtype


def test_dataset_config_dtype_raises():
    with pytest.raises(ValueError):
        XarrayDataConfig(data_path="path/to/data", dtype="invalid_dtype")


@pytest.mark.parametrize(
    "mock_data_fixture, engine, file_pattern",
    [
        ("mock_monthly_netcdfs_with_nans", "netcdf4", "*.nc"),
        ("mock_monthly_zarr_with_nans", "zarr", "*.zarr"),
    ],
)
def test_fill_nans(mock_data_fixture, engine, file_pattern, request):
    mock_data: MockData = request.getfixturevalue(mock_data_fixture)
    nan_config = FillNaNsConfig()
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir,
        fill_nans=nan_config,
        engine=engine,
        file_pattern=file_pattern,
    )
    names = mock_data.var_names.all_names
    dataset = XarrayDataset(config, names, 2)
    data, _ = dataset[0]
    assert torch.all(data["foo"][0, :, 0] == 0)
    assert torch.all(data["constant_var"][:, 0, 0] == 0)


def test_keep_nans(mock_monthly_netcdfs_with_nans):
    config_keep_nan = XarrayDataConfig(data_path=mock_monthly_netcdfs_with_nans.tmpdir)
    names = mock_monthly_netcdfs_with_nans.var_names.all_names
    dataset = XarrayDataset(config_keep_nan, names, 2)
    data_with_nan, _ = dataset[0]
    assert torch.all(torch.isnan(data_with_nan["foo"][0, :, 0]))
    assert torch.all(torch.isnan(data_with_nan["constant_var"][:, 0, 0]))


def test_overwrite(mock_monthly_netcdfs):
    const = -10
    multiple = 3.5

    overwrite_config = OverwriteConfig(
        constant={"foo": const},
        multiply_scalar={"bar": multiple},
    )

    config = XarrayDataConfig(data_path=mock_monthly_netcdfs.tmpdir)
    n_timesteps = 2
    names = mock_monthly_netcdfs.var_names.all_names
    dataset = XarrayDataset(config, names, n_timesteps)[0][0]

    config_overwrite = XarrayDataConfig(
        data_path=mock_monthly_netcdfs.tmpdir, overwrite=overwrite_config
    )
    n_timesteps = 2
    dataset_overwrite = XarrayDataset(config_overwrite, names, n_timesteps)[0][0]

    for v in ["foo", "bar"]:
        assert dataset_overwrite[v].dtype == dataset[v].dtype
        assert dataset_overwrite[v].device == dataset[v].device
    assert torch.equal(
        dataset_overwrite["foo"], torch.ones_like(dataset["foo"]) * const
    )
    assert torch.equal(dataset_overwrite["bar"], dataset["bar"] * multiple)


def test_repeated_interval_boolean_mask_subset(mock_monthly_netcdfs):
    config = XarrayDataConfig(data_path=mock_monthly_netcdfs.tmpdir)
    names = mock_monthly_netcdfs.var_names.all_names
    dataset = XarrayDataset(config, names, 1)
    interval = RepeatedInterval(interval_length="1D", block_length="7D", start="3D")
    boolean_mask = interval.get_boolean_mask(len(dataset), dataset.timestep)
    subset = XarraySubset(dataset, boolean_mask)

    # Check that the subset length matches the expected number of intervals
    expected_length = boolean_mask.sum().item()
    assert len(subset) == expected_length


def test_multi_source_xarray_dataset_has_no_duplicates(
    mock_monthly_netcdfs, mock_monthly_netcdfs_another_source
):
    monthly_netcdfs = [mock_monthly_netcdfs, mock_monthly_netcdfs_another_source]
    datasets = []

    for mock_data in monthly_netcdfs:
        config_source = XarrayDataConfig(data_path=mock_data.tmpdir)
        names = mock_data.var_names.all_names
        dataset = XarrayDataset(config_source, names, 1)
        datasets.append(dataset)

    with pytest.raises(ValueError):
        # duplicate variable names
        MergedXarrayDataset(datasets=datasets)


def test_multi_source_xarray_dataset_has_same_time(
    mock_monthly_netcdfs, mock_monthly_netcdfs_another_source_diff_time
):
    monthly_netcdfs = [
        mock_monthly_netcdfs,
        mock_monthly_netcdfs_another_source_diff_time,
    ]

    datasets = []
    for mock_data in monthly_netcdfs:
        config_source = XarrayDataConfig(data_path=mock_data.tmpdir)
        names = mock_data.var_names.all_names
        dataset = XarrayDataset(config_source, names, 1)
        datasets.append(dataset)
    # different time index
    with pytest.raises(ValueError):
        MergedXarrayDataset(datasets=datasets)


def test_multi_source_xarray_returns_merged_data(
    mock_monthly_netcdfs, mock_monthly_netcdfs_another_source
):
    config_source1 = XarrayDataConfig(data_path=mock_monthly_netcdfs.tmpdir)
    names1 = mock_monthly_netcdfs.var_names.all_names
    dataset1 = XarrayDataset(config_source1, names1, 1)

    config_source2 = XarrayDataConfig(
        data_path=mock_monthly_netcdfs_another_source.tmpdir
    )
    names2 = mock_monthly_netcdfs_another_source.var_names.all_names
    # remove duplicates in source 2 requirements
    for name in names1:
        if name in names2:
            names2.remove(name)
    dataset2 = XarrayDataset(config_source2, names2, 1)
    merged_dataset = MergedXarrayDataset(datasets=[dataset1, dataset2])
    assert len(merged_dataset) == len(dataset1)
    assert type(merged_dataset[0]) is type(dataset1[0])
    assert type(merged_dataset[0]) is type(dataset2[0])
    for key in merged_dataset[0][0].keys():
        if key in dataset1[0][0].keys():
            assert torch.equal(merged_dataset[0][0][key], dataset1[0][0][key])
            assert merged_dataset[0][1].equals(dataset1[0][1])
        if key in dataset2[0][0].keys():
            assert torch.equal(merged_dataset[0][0][key], dataset2[0][0][key])
            assert merged_dataset[0][1].equals(dataset2[0][1])
        else:
            assert KeyError(f"Key {key} is missing in merged dataset")


def test_xarray_subset_has_correct_sample(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    config = XarrayDataConfig(data_path=mock_data.tmpdir)
    config2 = XarrayDataConfig(data_path=mock_data.tmpdir, subset=Slice(stop=1))
    n_timesteps = 5
    names = mock_data.var_names.all_names + ["x"]
    dataset, _ = get_xarray_dataset(config, names, n_timesteps)
    dataset2, _ = get_xarray_dataset(config2, names, n_timesteps)
    assert dataset.sample_start_times[0:1].equals(dataset2.sample_start_times)
    assert dataset[0][0]["foo"].equal(dataset2[0][0]["foo"])
    assert dataset[0][1].equals(dataset2[0][1])


def test_xarray_concat_has_correct_sample(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    n_timesteps = 5
    names = mock_data.var_names.all_names + ["x"]
    config1 = XarrayDataConfig(
        data_path=mock_data.tmpdir, subset=TimeSlice("2003-03-01", "2003-03-31")
    )

    config2 = XarrayDataConfig(
        data_path=mock_data.tmpdir, subset=TimeSlice("2003-05-01", "2003-05-31")
    )
    concat, properties = get_dataset([config1, config2], names, n_timesteps)
    expected1, _ = get_xarray_dataset(config1, names, n_timesteps)
    expected2, _ = get_xarray_dataset(config2, names, n_timesteps)
    expected_times = np.concatenate(
        [expected1.sample_start_times, expected2.sample_start_times]
    )
    expected = xr.CFTimeIndex(expected_times)
    assert concat.sample_start_times.equals(expected)


def test__get_vertical_coordinate_raises():
    data = xr.Dataset({"ak_0": 1.0, "bk_0": 0.5, "idepth_0": 1.0})
    with pytest.raises(ValueError, match="Dataset contains both hybrid"):
        _get_vertical_coordinate(data, dtype=None)


def test__get_vertical_coordinate_null():
    data = xr.Dataset()
    vertical_coordinate = _get_vertical_coordinate(data, dtype=None)
    assert vertical_coordinate == NullVerticalCoordinate()


def test__get_vertical_coordinate_hybrid_sigma_pressure():
    data = xr.Dataset({"ak_0": 1.0, "bk_0": 0.5, "ak_1": 2.0, "bk_1": 1.5})
    vertical_coordinate = _get_vertical_coordinate(data, dtype=None)
    assert isinstance(vertical_coordinate, HybridSigmaPressureCoordinate)
    assert vertical_coordinate.ak[0] == 1.0
    assert vertical_coordinate.bk[0] == 0.5


def test__get_vertical_coordinate_depth_no_mask():
    data = xr.Dataset({"idepth_0": 1.0, "idepth_1": 2.0})
    vertical_coordinate = _get_vertical_coordinate(data, dtype=None)
    assert isinstance(vertical_coordinate, DepthCoordinate)
    assert vertical_coordinate.idepth[0] == 1.0
    assert vertical_coordinate.mask[0] == 1.0


def test__get_vertical_coordinate_depth_with_lat_dependent_mask():
    data = xr.Dataset(
        data_vars={
            "idepth_0": 1.0,
            "idepth_1": 2.0,
            "idepth_2": 3.0,
            "mask_0": ("lat", np.array([1.0, 1.0])),
            "mask_1": ("lat", np.array([0.0, 1.0])),
        },
        coords={
            "lat": np.array([1.0, 2.0]),
        },
    )
    vertical_coordinate = _get_vertical_coordinate(data, dtype=None)
    assert isinstance(vertical_coordinate, DepthCoordinate)
    assert vertical_coordinate.idepth[0] == 1.0
    assert vertical_coordinate.idepth.shape == (3,)
    assert vertical_coordinate.mask.shape == (2, 2)


def test__get_vertical_coordinate_depth_with_time_dependent_mask():
    data = xr.Dataset(
        data_vars={
            "idepth_0": 1.0,
            "idepth_1": 2.0,
            "idepth_2": 3.0,
            "mask_0": ("time", np.array([1.0, 1.0])),
            "mask_1": ("time", np.array([0.0, 1.0])),
        },
        coords={
            "time": np.array([1.0, 2.0]),
        },
    )
    with pytest.raises(ValueError, match="The ocean mask must by time-independent."):
        _get_vertical_coordinate(data, dtype=None)


@pytest.mark.parametrize(
    "kwargs,",
    [
        pytest.param({"spatial_dimensions": "xyz"}, id="invalid_spatial_dimensions"),
        pytest.param(
            {"engine": "zarr", "file_pattern": "*.nc"},
            id="engine_file_pattern_mismatch",
        ),
        pytest.param(
            {"n_repeats": 2, "infer_timestep": False}, id="n_repeats_infer_timestep"
        ),
        pytest.param({"dtype": "foo"}, id="invalid_dtype"),
    ],
)
def test_invalid_config_field_raises_error(kwargs):
    """Runs shape and length checks on the dataset."""
    with pytest.raises(ValueError):
        XarrayDataConfig(data_path="path", **kwargs)


@pytest.mark.parametrize(
    "mock_data_fixture, engine, file_pattern",
    [
        ("mock_monthly_netcdfs_ensemble_dim", "netcdf4", "*.nc"),
        ("mock_monthly_zarr_ensemble_dim", "zarr", "*.zarr"),
    ],
)
def test_dataset_with_nonspacetime_dim(
    mock_data_fixture, engine, file_pattern, request
):
    mock_data: MockData = request.getfixturevalue(mock_data_fixture)
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir,
        dtype="bfloat16",
        engine=engine,
        file_pattern=file_pattern,
    )
    # Omit the test variable that has mismatch dimensions
    vars = list(set(mock_data.var_names.all_names) - {"var_no_ensemble_dim"})
    dataset = XarrayDataset(config, vars, 2)
    data, _ = dataset[0]
    assert len(data["foo"].shape) == 4
    assert dataset.dims == ["time", "sample", "lat", "lon"]


@pytest.mark.parametrize(
    "mock_data_fixture, engine, file_pattern",
    [
        ("mock_monthly_netcdfs_ensemble_dim", "netcdf4", "*.nc"),
        ("mock_monthly_zarr_ensemble_dim", "zarr", "*.zarr"),
    ],
)
def test_dataset_raise_error_on_dim_mismatch(
    mock_data_fixture, engine, file_pattern, request
):
    # Should raise error when trying to broadcast variable that is missing
    # ensemble 'sample' dim
    mock_data: MockData = request.getfixturevalue(mock_data_fixture)
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir,
        dtype="bfloat16",
        engine=engine,
        file_pattern=file_pattern,
    )
    dataset = XarrayDataset(config, mock_data.var_names.all_names, 2)
    with pytest.raises(ValueError):
        dataset[0]


def test_xarray_raise_error_on_concat_dim_mismatch(
    mock_monthly_netcdfs, mock_monthly_netcdfs_ensemble_dim
):
    mock_data: MockData = mock_monthly_netcdfs
    mock_data_ensemble: MockData = mock_monthly_netcdfs_ensemble_dim
    n_timesteps = 5
    names = mock_data.var_names.all_names + ["x"]
    config1 = XarrayDataConfig(
        data_path=mock_data.tmpdir, subset=TimeSlice("2003-03-01", "2003-03-31")
    )
    config2 = XarrayDataConfig(
        data_path=mock_data_ensemble.tmpdir,
        subset=TimeSlice("2003-05-01", "2003-05-31"),
    )
    with pytest.raises(ValueError):
        get_dataset([config1, config2], names, n_timesteps)


@pytest.mark.parametrize(
    "mock_data_fixture, engine, file_pattern",
    [
        ("mock_monthly_netcdfs_ensemble_dim", "netcdf4", "*.nc"),
        ("mock_monthly_zarr_ensemble_dim", "zarr", "*.zarr"),
    ],
)
def test_xarray_dataset_isel(mock_data_fixture, engine, file_pattern, request):
    mock_data: MockData = request.getfixturevalue(mock_data_fixture)
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir,
        engine=engine,
        file_pattern=file_pattern,
        subset=Slice(start=None, stop=2),
        isel={"sample": 1},
    )
    vars = list(set(mock_data.var_names.all_names) - {"var_no_ensemble_dim"})
    dataset = XarrayDataset(config, vars, 2)
    data, _ = dataset[0]
    # Original lat/lon sizes are 4, 8
    assert data["var_matches_sample_index"].shape == (2, 4, 8)
    assert data["constant_var"].shape == (2, 4, 8)
    assert "sample" not in dataset.dims
    assert torch.all(data["var_matches_sample_index"] == 1.0)


@pytest.mark.parametrize(
    "isel",
    [
        {"lat": 0},
        {"time": 0},
        {"grid_x": 0},
    ],
)
def test_xarray_dataset_invalid_isel_raises_error(
    mock_monthly_netcdfs_ensemble_dim, isel
):
    mock_data: MockData = mock_monthly_netcdfs_ensemble_dim
    names = mock_data.var_names.all_names

    with pytest.raises(ValueError):
        config = XarrayDataConfig(
            data_path=mock_data.tmpdir,
            subset=TimeSlice("2003-03-01", "2003-03-31"),
            isel=isel,
        )
        get_dataset([config], names, 5)


@pytest.mark.parametrize(
    "isel_value",
    [3, Slice(3, 13)],
)
def test_XarrayDataset_error_on_isel_outside_data(
    mock_monthly_netcdfs_ensemble_dim, isel_value
):
    # mock data has sample dimension size 3
    mock_data: MockData = mock_monthly_netcdfs_ensemble_dim
    config = XarrayDataConfig(
        data_path=mock_data.tmpdir,
        subset=Slice(start=None, stop=2),
        isel={"sample": isel_value},
    )
    vars = list(set(mock_data.var_names.all_names) - {"var_no_ensemble_dim"})
    with pytest.raises(ValueError):
        XarrayDataset(config, vars, 2)


def test_concat_of_XarrayConcat(mock_monthly_netcdfs):
    mock_data: MockData = mock_monthly_netcdfs
    n_timesteps = 5
    names = mock_data.var_names.all_names + ["x"]
    config = XarrayDataConfig(data_path=mock_data.tmpdir, subset=Slice(None, 4))
    concat, _ = get_dataset([config, config], names, n_timesteps)
    concat2 = XarrayConcat(datasets=[concat, concat])
    assert len(concat2) == 16



import datetime
import unittest.mock
from collections import namedtuple
from collections.abc import Callable, Iterable

import numpy as np
import pytest
import torch
import xarray as xr

import fme
from fme.ace.data_loading.batch_data import BatchData, PrognosticState
from fme.ace.stepper import SingleModuleStepperConfig
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.generics.data import SimpleInferenceData
from fme.core.generics.inference import (
    Looper,
    PredictFunction,
    get_record_to_wandb,
    run_inference,
)
from fme.core.loss import WeightedMappingLossConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.registry.module import ModuleSelector
from fme.core.testing.wandb import mock_wandb
from fme.core.timing import GlobalTimer
from fme.core.typing_ import TensorDict, TensorMapping

SphericalData = namedtuple("SphericalData", ["data", "area_weights", "vertical_coord"])


def get_data(
    names: Iterable[str], shape: tuple[int, int, int, int, int]
) -> SphericalData:
    data = {}
    n_lat = shape[2]
    n_lon = shape[3]
    shape_without_z = shape[:-1]
    nz = shape[-1]
    lats = torch.linspace(-89.5, 89.5, n_lat)
    for name in names:
        data[name] = torch.rand(*shape_without_z, device=fme.get_device())
    area_weights = fme.spherical_area_weights(lats, n_lon).to(fme.get_device())
    ak, bk = torch.arange(nz), torch.arange(nz)
    vertical_coord = HybridSigmaPressureCoordinate(ak, bk)
    return SphericalData(data, area_weights, vertical_coord)


def get_scalar_data(names, value):
    return {n: value for n in names}


class MockLoader(torch.utils.data.DataLoader):
    def __init__(
        self,
        shape: tuple,
        names: Iterable[str],
        n_windows: int,
        time: xr.DataArray | None = None,
    ):
        device = fme.get_device()
        self._data = {n: torch.rand(*shape, device=device) for n in names}
        if time is None:
            self._time = xr.DataArray(np.zeros(shape[:2]), dims=["sample", "time"])
        elif time.shape != shape[:2]:
            raise ValueError(
                "Time shape must match the first two dimensions of the data."
            )
        else:
            self._time = time
        self._n_windows = n_windows
        self._current_window = 0

    def __iter__(self):
        return self

    def __len__(self) -> int:
        return self._n_windows

    def __next__(self) -> BatchData:
        if self._current_window < self._n_windows:
            self._current_window += 1
            return BatchData.new_on_device(
                data=self._data,
                time=self._time
                + (self._current_window - 1) * (self._time.shape[1] - 1),
            )
        else:
            raise StopIteration


def _get_stepper():
    class ChannelSum(torch.nn.Module):
        def forward(self, x):
            summed = torch.sum(x, dim=-3, keepdim=True)
            diagnostic = torch.rand_like(summed)
            return torch.concat([x, diagnostic], dim=-3)

    n_samples = 2
    n_time = 5
    n_lat = 2
    n_lon = 4
    nz = 3
    shape = (n_samples, n_time, n_lat, n_lon, nz)
    in_names = ["forcing", "prognostic"]
    out_names = ["prognostic", "diagnostic"]
    all_names = list(set(in_names + out_names))

    spherical_data = get_data(all_names, shape)
    time = xr.DataArray(
        np.repeat(np.expand_dims(np.arange(n_time), axis=0), n_samples, axis=0),
        dims=["sample", "time"],
    )

    img_shape = spherical_data.data[in_names[0]].shape[2:]
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(img_shape[0]),
        lon=torch.zeros(img_shape[1]),
    )
    vertical_coordinate = spherical_data.vertical_coord
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": ChannelSum()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means=get_scalar_data(all_names, 0.0),
            stds=get_scalar_data(all_names, 1.0),
        ),
        loss=WeightedMappingLossConfig(),
    )
    stepper = config.get_stepper(
        dataset_info=DatasetInfo(
            horizontal_coordinates=horizontal_coordinate,
            vertical_coordinate=vertical_coordinate,
            timestep=datetime.timedelta(seconds=1),
        ),
    )
    return stepper, spherical_data, time, in_names, out_names


def test_looper():
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    forcing_names = set(in_names) - set(out_names)
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, forcing_names, 3, time=time)
    looper = Looper(
        predict=stepper.predict,
        data=SimpleInferenceData(initial_condition, loader),
    )

    expected_output_shape = (shape[0], shape[1] - 1, shape[2], shape[3])
    for batch in looper:
        assert set(out_names) == set(batch.data)
        for name in out_names:
            assert batch.data[name].shape == expected_output_shape


def test_looper_paired():
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    forcing_names = set(in_names) - set(out_names)
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, forcing_names, 3, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    expected_output_shape = (shape[0], shape[1] - 1, shape[2], shape[3])
    for batch in looper:
        assert set(out_names) == set(batch.prediction)
        assert set(forcing_names) == set(batch.reference)
        for name in out_names:
            assert batch.prediction[name].shape == expected_output_shape
        for name in forcing_names:
            assert batch.reference[name].shape == expected_output_shape


def _mock_compute_derived_quantities(data, forcing_data):
    data_name = list(data)[0]
    forcing_name = list(forcing_data)[0]
    derived = {"derived": data[data_name] + forcing_data[forcing_name]}
    return {**data, **derived}


def test_looper_paired_with_derived_variables():
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=_mock_compute_derived_quantities
    )
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    stepper._derive_func = mock_derive_func
    forcing_names = set(in_names) - set(out_names)
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        {n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, forcing_names, 2, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    for batch in looper:
        assert "derived" in batch.prediction
        assert "derived" in batch.reference
    mock_derive_func.assert_called()


def test_looper_paired_with_target_data():
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    all_names = list(set(in_names + out_names))
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, all_names, 2, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    for batch in looper:
        assert set(out_names) == set(batch.prediction)
        assert set(all_names) == set(batch.reference)


def test_looper_paired_with_target_data_and_derived_variables():
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=_mock_compute_derived_quantities
    )
    stepper, spherical_data, time, in_names, out_names = _get_stepper()
    stepper._derive_func = mock_derive_func
    all_names = list(set(in_names + out_names))
    shape = spherical_data.data[in_names[0]].shape
    initial_condition = BatchData.new_on_device(
        data={n: spherical_data.data[n][:, :1] for n in spherical_data.data},
        time=time[:, 0:1],
    ).get_start(
        prognostic_names=stepper.prognostic_names,
        n_ic_timesteps=1,
    )
    loader = MockLoader(shape, all_names, 2, time=time)
    looper = Looper(
        predict=stepper.predict_paired,
        data=SimpleInferenceData(initial_condition, loader),
    )

    for batch in looper:
        assert set(out_names + ["derived"]) == set(batch.prediction)
        assert set(all_names + ["derived"]) == set(batch.reference)
    mock_derive_func.assert_called()


def get_batch_data(
    start_time,
    n_timesteps,
):
    n_samples = 1
    n_lat = 3
    n_lon = 4
    time_values = torch.arange(
        start_time, start_time + n_timesteps, device=get_device()
    )[None, :, None, None]
    time_axis = torch.broadcast_to(
        start_time + torch.arange(n_timesteps)[None, :], (n_samples, n_timesteps)
    )
    time = xr.DataArray(time_axis, dims=["sample", "time"])
    return BatchData.new_on_device(
        data={
            "var": torch.broadcast_to(
                time_values, (n_samples, n_timesteps, n_lat, n_lon)
            )
        },
        time=time,
    )


class PlusOneStepper:
    def __init__(
        self,
        n_ic_timesteps: int,
        derive_func: Callable[[TensorMapping, TensorMapping], TensorDict] | None = None,
    ):
        self.n_ic_timesteps = n_ic_timesteps
        if derive_func is None:
            self.derive_func: Callable[[TensorMapping, TensorMapping], TensorDict] = (
                unittest.mock.MagicMock(side_effect=lambda x, y=None: dict(x))
            )
        else:
            self.derive_func = derive_func
        _: PredictFunction[  # for type checking
            PrognosticState,
            BatchData,
            BatchData,
        ] = self.predict

    def predict(
        self,
        initial_condition: PrognosticState,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> tuple[BatchData, PrognosticState]:
        ic_state = initial_condition.as_batch_data()
        n_forward_steps = forcing.time.shape[1] - self.n_ic_timesteps
        out_tensor = torch.zeros(
            ic_state.data["var"].shape[0],
            n_forward_steps,
            *ic_state.data["var"].shape[2:],
            device=ic_state.data["var"].device,
            dtype=ic_state.data["var"].dtype,
        )
        out_tensor[:, 0, ...] = ic_state.data["var"][:, -1, ...] + 1
        for i in range(1, n_forward_steps):
            out_tensor[:, i, ...] = out_tensor[:, i - 1, ...] + 1
        data = BatchData.new_on_device(
            data={"var": out_tensor},
            time=forcing.time[:, self.n_ic_timesteps :],
        )
        if compute_derived_variables:
            data = data.compute_derived_variables(
                derive_func=self.derive_func, forcing_data=data
            )
        return data, data.get_end(["var"], self.n_ic_timesteps)

    def get_forward_data(
        self,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> BatchData:
        if compute_derived_variables:
            forcing = forcing.compute_derived_variables(
                derive_func=self.derive_func, forcing_data=forcing
            )
        return forcing.remove_initial_condition(self.n_ic_timesteps)


def test_looper_simple_batch_data():
    n_ic_timesteps = 1
    n_forward_steps = 2
    n_iterations = 10
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=lambda batch_data, forcing_data: batch_data
    )
    stepper = PlusOneStepper(
        n_ic_timesteps=n_ic_timesteps, derive_func=mock_derive_func
    )
    initial_condition = get_batch_data(
        0,
        n_timesteps=n_ic_timesteps,
    ).get_start(prognostic_names=["var"], n_ic_timesteps=n_ic_timesteps)
    loader = [
        get_batch_data(
            i,
            n_ic_timesteps + n_forward_steps,
        )
        for i in range(0, n_iterations * n_forward_steps, n_forward_steps)
    ]

    mock_predict = unittest.mock.MagicMock(side_effect=stepper.predict)
    with unittest.mock.patch.object(stepper, "predict", mock_predict):
        with GlobalTimer():
            timer = GlobalTimer.get_instance()
            looper = Looper(
                predict=stepper.predict,
                data=SimpleInferenceData(initial_condition, loader),
            )
            for i, batch in enumerate(looper):
                for j in range(batch.time.shape[1]):
                    assert torch.allclose(
                        batch.data["var"][:, j, ...],
                        torch.as_tensor(n_ic_timesteps + i * n_forward_steps + j),
                    )
            times = timer.get_durations()
            assert times["data_loading"] > 0
            assert mock_derive_func.call_count == n_iterations
            assert mock_predict.call_count == n_iterations
            # we mocked out the implicit calls that happen in .predict
            # if this changed, update the test
            assert "forward_prediction" not in times
            assert "compute_derived_variables" not in times


def get_mock_aggregator(
    n_ic_timesteps: int,
) -> unittest.mock.MagicMock:
    mock_aggregator = unittest.mock.MagicMock()

    # record_batch will start at step n_ic_timesteps
    i = n_ic_timesteps

    def record_batch_side_effect(
        data: BatchData,
    ):
        nonlocal i
        ret = [{"step": j} for j in range(i, i + data.time.shape[1])]
        i += data.time.shape[1]
        return ret

    mock_aggregator = unittest.mock.MagicMock()
    mock_aggregator.record_initial_condition = unittest.mock.MagicMock(
        return_value=[{"step": j} for j in range(n_ic_timesteps)]
    )

    def get_summary_logs_side_effect():
        # we expect this gets called _outside_ of run_inference
        raise ValueError("should not be called")

    mock_aggregator.get_summary_logs = unittest.mock.MagicMock(
        side_effect=get_summary_logs_side_effect
    )
    mock_aggregator.record_batch = unittest.mock.MagicMock(
        side_effect=record_batch_side_effect
    )
    return mock_aggregator


def get_mock_writer() -> unittest.mock.MagicMock:
    mock_writer = unittest.mock.MagicMock()
    return mock_writer


@pytest.mark.parametrize(
    "n_ic_timesteps, n_forward_steps, n_iterations",
    [
        pytest.param(1, 2, 5, id="n_ic_timesteps=1"),
        pytest.param(2, 2, 5, id="n_ic_timesteps=2"),
    ],
)
def test_run_inference_simple(
    n_ic_timesteps: int, n_forward_steps: int, n_iterations: int
):
    mock_derive_func = unittest.mock.MagicMock(
        side_effect=lambda batch_data, forcing_data: batch_data
    )
    stepper = PlusOneStepper(
        n_ic_timesteps=n_ic_timesteps, derive_func=mock_derive_func
    )
    initial_condition = get_batch_data(
        0,
        n_timesteps=n_ic_timesteps,
    ).get_start(prognostic_names=["var"], n_ic_timesteps=n_ic_timesteps)
    loader = [
        get_batch_data(
            i,
            n_ic_timesteps + n_forward_steps,
        )
        for i in range(0, n_iterations * n_forward_steps, n_forward_steps)
    ]
    mock_writer = get_mock_writer()
    mock_aggregator = get_mock_aggregator(n_ic_timesteps)

    with GlobalTimer():
        with mock_wandb() as wandb:
            wandb.configure(log_to_wandb=True)
            record_logs = unittest.mock.MagicMock(
                side_effect=get_record_to_wandb("inference")
            )  # this init must be within mock_wandb context
            run_inference(
                predict=stepper.predict,
                data=SimpleInferenceData(initial_condition, loader),
                writer=mock_writer,
                aggregator=mock_aggregator,
                record_logs=record_logs,
            )
            wandb_logs = wandb.get_logs()
        timer = GlobalTimer.get_instance()
        times = timer.get_durations()
        assert times["wandb_logging"] > 0
        assert times["data_writer"] > 0
        assert times["aggregator"] > 0
        assert mock_writer.write.call_count == 2
        assert mock_aggregator.record_initial_condition.call_count == 1
        assert mock_writer.append_batch.call_count == n_iterations
        assert mock_aggregator.record_batch.call_count == n_iterations
        assert len(wandb_logs) == n_ic_timesteps + n_iterations * n_forward_steps
        assert wandb_logs == [
            {"inference/step": i}
            for i in range(n_ic_timesteps + n_iterations * n_forward_steps)
        ]
        assert (
            record_logs.call_count == n_iterations + 1
        )  # +1 for the initial condition



import dataclasses
import datetime
import logging
import pathlib
import warnings
from collections.abc import Callable, Generator, Mapping
from typing import Any, Literal, cast

import dacite
import dacite.exceptions
import torch
import xarray as xr
from torch import nn

from fme.ace.data_loading.batch_data import BatchData, PairedData, PrognosticState
from fme.ace.requirements import DataRequirements, PrognosticStateDataRequirements
from fme.ace.stepper.parameter_init import (
    ParameterInitializationConfig,
    ParameterInitializer,
)
from fme.core.coordinates import (
    NullPostProcessFn,
    SerializableVerticalCoordinate,
    VerticalCoordinate,
)
from fme.core.corrector.atmosphere import AtmosphereCorrectorConfig
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset.utils import encode_timestep
from fme.core.dataset_info import DatasetInfo, MissingDatasetInfo
from fme.core.device import get_device
from fme.core.generics.inference import PredictFunction
from fme.core.generics.optimization import OptimizationABC
from fme.core.generics.train_stepper import TrainOutputABC, TrainStepperABC
from fme.core.loss import WeightedMappingLoss, WeightedMappingLossConfig
from fme.core.masking import NullMasking, StaticMaskingConfig
from fme.core.multi_call import MultiCallConfig
from fme.core.normalizer import (
    NetworkAndLossNormalizationConfig,
    NormalizationConfig,
    StandardNormalizer,
)
from fme.core.ocean import OceanConfig
from fme.core.optimization import NullOptimization
from fme.core.registry import CorrectorSelector, ModuleSelector
from fme.core.step.multi_call import MultiCallStepConfig, replace_multi_call
from fme.core.step.single_module import SingleModuleStepConfig
from fme.core.step.step import StepABC, StepSelector
from fme.core.tensors import (
    add_ensemble_dim,
    fold_ensemble_dim,
    fold_sized_ensemble_dim,
    repeat_interleave_batch_dim,
    unfold_ensemble_dim,
)
from fme.core.timing import GlobalTimer
from fme.core.training_history import TrainingHistory, TrainingJob
from fme.core.typing_ import EnsembleTensorDict, TensorDict, TensorMapping

DEFAULT_TIMESTEP = datetime.timedelta(hours=6)
DEFAULT_ENCODED_TIMESTEP = encode_timestep(DEFAULT_TIMESTEP)

Weights = list[Mapping[str, Any]]
StepperWeightsAndHistory = tuple[Weights, TrainingHistory]


def _load_weights_and_history(path: str) -> StepperWeightsAndHistory:
    stepper = load_stepper(path)
    return_weights: Weights = []
    for module in stepper.modules:
        return_weights.append(module.state_dict())
    return return_weights, stepper.training_history


@dataclasses.dataclass
class SingleModuleStepperConfig:
    """
    Configuration for a single module stepper.

    Parameters:
        builder: The module builder.
        in_names: Names of input variables.
        out_names: Names of output variables.
        normalization: The normalization configuration.
        parameter_init: The parameter initialization configuration.
        ocean: The ocean configuration.
        loss: The loss configuration.
        corrector: The corrector configuration.
        next_step_forcing_names: Names of forcing variables for the next timestep.
        loss_normalization: The normalization configuration for the loss.
        residual_normalization: Optional alternative to configure loss normalization.
            If provided, it will be used for all *prognostic* variables in loss scaling.
        multi_call: The configuration of multi-called diagnostics.
        include_multi_call_in_loss: Whether to include multi-call diagnostics in the
            loss. The same loss configuration as specified in 'loss' is used.
        crps_training: Whether to use CRPS training for stochastic models.
        residual_prediction: Whether to have ML module predict tendencies for
            prognostic variables.
    """

    builder: ModuleSelector
    in_names: list[str]
    out_names: list[str]
    normalization: NormalizationConfig
    parameter_init: ParameterInitializationConfig = dataclasses.field(
        default_factory=lambda: ParameterInitializationConfig()
    )
    ocean: OceanConfig | None = None
    loss: WeightedMappingLossConfig = dataclasses.field(
        default_factory=lambda: WeightedMappingLossConfig()
    )
    corrector: AtmosphereCorrectorConfig | CorrectorSelector = dataclasses.field(
        default_factory=lambda: AtmosphereCorrectorConfig()
    )
    next_step_forcing_names: list[str] = dataclasses.field(default_factory=list)
    loss_normalization: NormalizationConfig | None = None
    residual_normalization: NormalizationConfig | None = None
    multi_call: MultiCallConfig | None = None
    include_multi_call_in_loss: bool = False
    crps_training: bool = False
    residual_prediction: bool = False

    def __post_init__(self):
        for name in self.next_step_forcing_names:
            if name not in self.in_names:
                raise ValueError(
                    f"next_step_forcing_name '{name}' not in in_names: {self.in_names}"
                )
            if name in self.out_names:
                raise ValueError(
                    f"next_step_forcing_name is an output variable: '{name}'"
                )
        if (
            self.residual_normalization is not None
            and self.loss_normalization is not None
        ):
            raise ValueError(
                "Only one of residual_normalization, loss_normalization can "
                "be provided."
                "If residual_normalization is provided, it will be used for all "
                "*prognostic* variables in loss scalng. "
                "If loss_normalization is provided, it will be used for all variables "
                "in loss scaling."
            )
        if self.multi_call is not None:
            self.multi_call.validate(self.in_names, self.out_names)
        if self.include_multi_call_in_loss:
            if self.multi_call is None:
                raise ValueError(
                    "include_multi_calls_in_loss is True but no multi_call config "
                    "was provided."
                )

    def load(self):
        self.normalization.load()
        if self.loss_normalization is not None:
            self.loss_normalization.load()
        if self.residual_normalization is not None:
            self.residual_normalization.load()

    @property
    def n_ic_timesteps(self) -> int:
        return 1

    def get_evaluation_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return DataRequirements(
            names=self.all_names,
            n_timesteps=self._window_steps_required(n_forward_steps),
        )

    def get_prognostic_state_data_requirements(self) -> PrognosticStateDataRequirements:
        return PrognosticStateDataRequirements(
            names=self.prognostic_names,
            n_timesteps=self.n_ic_timesteps,
        )

    def _window_steps_required(self, n_forward_steps: int) -> int:
        return n_forward_steps + self.n_ic_timesteps

    def get_state(self):
        self.load()
        return dataclasses.asdict(self)

    def get_parameter_initializer(self) -> ParameterInitializer:
        """
        Get the parameter initializer for this stepper configuration.
        """
        return self.parameter_init.build(
            load_weights_and_history=_load_weights_and_history
        )

    def get_stepper(
        self,
        dataset_info: DatasetInfo,
        apply_parameter_init: bool = True,
    ) -> "Stepper":
        """
        Args:
            dataset_info: Information about the training dataset.
            apply_parameter_init: Whether to apply parameter initialization.
        """
        logging.info("Initializing stepper from provided legacy config")
        normalizer = self.normalization.build(self.normalize_names)
        combined_normalization_config = NetworkAndLossNormalizationConfig(
            network=self.normalization,
            loss=self.loss_normalization,
            residual=self.residual_normalization,
        )
        loss_normalizer = combined_normalization_config.get_loss_normalizer(
            self.normalize_names, residual_scaled_names=self.prognostic_names
        )
        new_config = self.to_stepper_config(
            normalizer=normalizer, loss_normalizer=loss_normalizer
        )
        return new_config.get_stepper(
            dataset_info=dataset_info,
            apply_parameter_init=apply_parameter_init,
        )

    def get_ocean(self) -> OceanConfig | None:
        return self.ocean

    @classmethod
    def from_state(cls, state) -> "SingleModuleStepperConfig":
        state = cls.remove_deprecated_keys(state)
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )

    @property
    def input_names(self) -> list[str]:
        return self.in_names

    @property
    def output_names(self) -> list[str]:
        return self.out_names

    @property
    def all_names(self):
        """Names of all variables required, including auxiliary ones."""
        extra_names = []
        if self.ocean is not None:
            extra_names.extend(self.ocean.forcing_names)
        if self.multi_call is not None:
            extra_names.extend(self.multi_call.names)
        all_names = list(set(self.in_names).union(self.out_names).union(extra_names))
        return all_names

    @property
    def normalize_names(self):
        """Names of variables which require normalization. I.e. inputs/outputs."""
        extra_names = []
        if self.multi_call is not None:
            extra_names.extend(self.multi_call.names)
        return list(set(self.in_names).union(self.out_names).union(extra_names))

    @property
    def input_only_names(self) -> list[str]:
        """Names of variables which are inputs only."""
        return list(set(self.all_names) - set(self.out_names))

    @property
    def prognostic_names(self) -> list[str]:
        """Names of variables which both inputs and outputs."""
        return list(set(self.out_names).intersection(self.in_names))

    @property
    def loss_names(self) -> list[str]:
        extra_names = []
        if self.multi_call is not None:
            extra_names.extend(self.multi_call.names)
        return list(set(self.out_names).union(extra_names))

    @property
    def diagnostic_names(self) -> list[str]:
        """Names of variables which are outputs only."""
        extra_names = []
        if self.multi_call is not None:
            extra_names = self.multi_call.names
        out_names = list(set(self.out_names).union(extra_names))
        return list(set(out_names).difference(self.in_names))

    @classmethod
    def remove_deprecated_keys(cls, state: dict[str, Any]) -> dict[str, Any]:
        _unsupported_key_defaults = {
            "conserve_dry_air": False,
            "optimization": None,
            "conservation_loss": {"dry_air_penalty": None},
        }
        state_copy = state.copy()
        for key, default in _unsupported_key_defaults.items():
            if key in state_copy:
                if state_copy[key] == default or state_copy[key] is None:
                    del state_copy[key]
                else:
                    raise ValueError(
                        f"The stepper config option {key} is deprecated and the setting"
                        f" provided, {state_copy[key]}, is no longer implemented. The "
                        "SingleModuleStepper being loaded from state cannot be run by "
                        "this version of the code."
                    )
        for normalization_key in [
            "normalization",
            "loss_normalization",
            "residual_normalization",
        ]:
            if state_copy.get(normalization_key) is not None:
                if "exclude_names" in state_copy[normalization_key]:
                    if state_copy[normalization_key]["exclude_names"] is not None:
                        raise ValueError(
                            "The exclude_names option in normalization config is no "
                            "longer supported, but excluded names were found in "
                            f"{normalization_key}."
                        )
                    else:
                        del state_copy[normalization_key]["exclude_names"]
        if "prescriber" in state_copy:
            # want to maintain backwards compatibility for this particular feature
            if state_copy["prescriber"] is not None:
                if state_copy.get("ocean") is not None:
                    raise ValueError("Cannot specify both prescriber and ocean.")
                state_copy["ocean"] = {
                    "surface_temperature_name": state_copy["prescriber"][
                        "prescribed_name"
                    ],
                    "ocean_fraction_name": state_copy["prescriber"]["mask_name"],
                    "interpolate": state_copy["prescriber"]["interpolate"],
                }
            del state_copy["prescriber"]
        if "activation_checkpointing" in state_copy:
            del state_copy["activation_checkpointing"]
        return state_copy

    def to_stepper_config(
        self,
        normalizer: StandardNormalizer,
        loss_normalizer: StandardNormalizer,
    ) -> "StepperConfig":
        """
        Convert the current config to a stepper config.

        Overwriting normalization configuration is needed to avoid
        a checkpoint trying to load normalization data from netCDF files
        which are no longer present when running inference.

        Args:
            normalizer: overwrite the normalization config
                with data from this normalizer
            loss_normalizer: overwrite the loss normalization config
                with data from this normalizer

        Returns:
            A stepper config.
        """
        return StepperConfig(
            step=self._to_step_config(normalizer, loss_normalizer),
            loss=self.loss,
            crps_training=self.crps_training,
            parameter_init=self.parameter_init,
        )

    def _to_step_config(
        self,
        normalizer: StandardNormalizer | None = None,
        loss_normalizer: StandardNormalizer | None = None,
    ) -> StepSelector:
        return StepSelector(
            type="multi_call",
            config=dataclasses.asdict(
                MultiCallStepConfig(
                    wrapped_step=StepSelector(
                        type="single_module",
                        config=dataclasses.asdict(
                            self._to_single_module_step_config(
                                normalizer=normalizer,
                                loss_normalizer=loss_normalizer,
                            )
                        ),
                    ),
                    config=self.multi_call,
                    include_multi_call_in_loss=self.include_multi_call_in_loss,
                )
            ),
        )

    def _to_single_module_step_config(
        self,
        normalizer: StandardNormalizer | None = None,
        loss_normalizer: StandardNormalizer | None = None,
    ) -> "SingleModuleStepConfig":
        if normalizer is not None:
            normalization = normalizer.get_normalization_config()
        else:
            normalization = self.normalization
        if loss_normalizer is not None:
            loss_normalization: NormalizationConfig | None = (
                loss_normalizer.get_normalization_config()
            )
            residual_normalization: NormalizationConfig | None = None
        else:
            loss_normalization = self.loss_normalization
            residual_normalization = self.residual_normalization
        return SingleModuleStepConfig(
            builder=self.builder,
            in_names=self.in_names,
            out_names=self.out_names,
            normalization=NetworkAndLossNormalizationConfig(
                network=normalization,
                loss=loss_normalization,
                residual=residual_normalization,
            ),
            ocean=self.ocean,
            corrector=self.corrector,
            next_step_forcing_names=self.next_step_forcing_names,
            crps_training=self.crps_training,
            residual_prediction=self.residual_prediction,
        )

    def replace_multi_call(self, multi_call: MultiCallConfig | None):
        self.multi_call = multi_call

    def replace_ocean(self, ocean: OceanConfig | None):
        self.ocean = ocean


@dataclasses.dataclass
class ExistingStepperConfig:
    """
    Configuration for an existing stepper. This allows loading a serialized
    stepper from a checkpoint without loading its configuration of the training
    and optimization schedule, i.e., this allows for specifying a new
    schedule in fine-tuning. Not used for training resumption.

    Parameters:
        checkpoint_path: The path to the serialized checkpoint; should be different
        than the experiment output directory.
    """

    checkpoint_path: str

    def __post_init__(self):
        self._stepper_config = StepperConfig.from_stepper_state(
            self._load_checkpoint()["stepper"]
        )

    def _load_checkpoint(self) -> Mapping[str, Any]:
        return torch.load(
            self.checkpoint_path, map_location=get_device(), weights_only=False
        )

    def get_evaluation_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return self._stepper_config.get_evaluation_window_data_requirements(
            n_forward_steps
        )

    def get_prognostic_state_data_requirements(self) -> PrognosticStateDataRequirements:
        return self._stepper_config.get_prognostic_state_data_requirements()

    def get_forcing_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return self._stepper_config.get_forcing_window_data_requirements(
            n_forward_steps
        )

    def get_parameter_initializer(self) -> ParameterInitializer:
        """Get a parameter initializer for this stepper configuration."""
        return self._stepper_config.get_parameter_initializer()

    def get_stepper(
        self,
        dataset_info: DatasetInfo,
        apply_parameter_init: bool = True,
    ):
        logging.info(f"Initializing stepper from {self.checkpoint_path}")
        return Stepper.from_state(self._load_checkpoint()["stepper"])


def _prepend_timesteps(
    data: EnsembleTensorDict, timesteps: TensorMapping, time_dim: int = 2
) -> EnsembleTensorDict:
    for v in data.values():
        n_ensemble = v.shape[1]
        break
    else:
        return data  # data is length zero
    timesteps = add_ensemble_dim(timesteps, repeats=n_ensemble)
    return EnsembleTensorDict(
        {k: torch.cat([timesteps[k], v], dim=time_dim) for k, v in data.items()}
    )


@dataclasses.dataclass
class TrainOutput(TrainOutputABC):
    metrics: TensorDict
    gen_data: EnsembleTensorDict
    target_data: EnsembleTensorDict
    time: xr.DataArray
    normalize: Callable[[TensorDict], TensorDict]
    derive_func: Callable[[TensorMapping, TensorMapping], TensorDict] = (
        lambda x, _: dict(x)
    )

    def __post_init__(self):
        for v in self.target_data.values():
            if v.shape[1] != 1:
                raise ValueError(
                    f"target_data can only have one ensemble member, got {v.shape[1]}"
                )

    def ensemble_derive_func(
        self, data: EnsembleTensorDict, forcing_data: TensorMapping
    ) -> EnsembleTensorDict:
        flattened_data, n_ensemble = fold_ensemble_dim(data)
        if n_ensemble > 1:
            ensemble_forcing_data = add_ensemble_dim(forcing_data, repeats=n_ensemble)
            flattened_forcing_data = fold_sized_ensemble_dim(
                ensemble_forcing_data, n_ensemble
            )
        else:
            flattened_forcing_data = dict(forcing_data)
        derived_data = self.derive_func(flattened_data, flattened_forcing_data)
        return unfold_ensemble_dim(derived_data, n_ensemble)

    def remove_initial_condition(self, n_ic_timesteps: int) -> "TrainOutput":
        return TrainOutput(
            metrics=self.metrics,
            gen_data=EnsembleTensorDict(
                {k: v[:, :, n_ic_timesteps:] for k, v in self.gen_data.items()}
            ),
            target_data=EnsembleTensorDict(
                {k: v[:, :, n_ic_timesteps:] for k, v in self.target_data.items()}
            ),
            time=self.time[:, n_ic_timesteps:],
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def copy(self) -> "TrainOutput":
        """Creates new dictionaries for the data but with the same tensors."""
        return TrainOutput(
            metrics=self.metrics,
            gen_data=EnsembleTensorDict({k: v for k, v in self.gen_data.items()}),
            target_data=EnsembleTensorDict({k: v for k, v in self.target_data.items()}),
            time=self.time,
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def prepend_initial_condition(
        self,
        initial_condition: PrognosticState,
    ) -> "TrainOutput":
        """
        Prepends an initial condition to the existing stepped data.
        Assumes data are on the same device.
        For data windows > 0, the target IC is different from the generated IC
            and may be provided for correct calculation of tendencies.

        Args:
            initial_condition: Initial condition data.
        """
        batch_data = initial_condition.as_batch_data()
        return TrainOutput(
            metrics=self.metrics,
            gen_data=_prepend_timesteps(self.gen_data, batch_data.data),
            target_data=_prepend_timesteps(
                self.target_data,
                batch_data.data,
            ),
            time=xr.concat([batch_data.time, self.time], dim="time"),
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def compute_derived_variables(
        self,
    ) -> "TrainOutput":
        gen_data = self.ensemble_derive_func(
            self.gen_data, fold_sized_ensemble_dim(self.target_data, 1)
        )
        target_data = self.ensemble_derive_func(
            self.target_data, fold_sized_ensemble_dim(self.target_data, 1)
        )
        return TrainOutput(
            metrics=self.metrics,
            gen_data=gen_data,
            target_data=target_data,
            time=self.time,
            normalize=self.normalize,
            derive_func=self.derive_func,
        )

    def get_metrics(self) -> TensorDict:
        return self.metrics


def stack_list_of_tensor_dicts(
    dict_list: list[TensorDict],
    time_dim: int,
) -> TensorDict:
    keys = next(iter(dict_list)).keys()
    stack_dict = {}
    for k in keys:
        stack_dict[k] = torch.stack([d[k] for d in dict_list], dim=time_dim)
    return stack_dict


def process_ensemble_prediction_generator_list(
    output_list: list[EnsembleTensorDict],
) -> EnsembleTensorDict:
    output_timeseries = stack_list_of_tensor_dicts(
        cast(list[TensorDict], output_list), time_dim=2
    )
    return EnsembleTensorDict(
        {k: v for k, v in output_timeseries.items()},
    )


def process_prediction_generator_list(
    output_list: list[TensorDict],
    time: xr.DataArray,
    horizontal_dims: list[str] | None = None,
) -> BatchData:
    output_timeseries = stack_list_of_tensor_dicts(output_list, time_dim=1)
    return BatchData.new_on_device(
        data=output_timeseries,
        time=time,
        horizontal_dims=horizontal_dims,
    )


@dataclasses.dataclass
class StepperConfig:
    """
    Configuration for a stepper.

    Parameters:
        step: The step configuration.
        loss: The loss configuration.
        n_ensemble: The number of ensemble members evaluated for each training
            batch member. Default is 2 if the loss type is EnsembleLoss, otherwise
            the default is 1. Must be 2 for EnsembleLoss to be valid.
        crps_training: Deprecated, kept for backwards compatibility. Use
            n_ensemble=2 with a CRPS loss instead.
        parameter_init: The parameter initialization configuration.
        input_masking: Config for masking step inputs.
    """

    step: StepSelector
    loss: WeightedMappingLossConfig = dataclasses.field(
        default_factory=lambda: WeightedMappingLossConfig()
    )
    n_ensemble: int = -1  # sentinel value to avoid None typing of attribute
    crps_training: bool = False
    parameter_init: ParameterInitializationConfig = dataclasses.field(
        default_factory=lambda: ParameterInitializationConfig()
    )
    input_masking: StaticMaskingConfig | None = None

    def __post_init__(self):
        if self.crps_training:
            warnings.warn(
                "crps_training is deprecated, use n_ensemble=2 "
                "with a CRPS loss instead",
                DeprecationWarning,
            )
            self.n_ensemble = 2
            self.loss = WeightedMappingLossConfig(
                type="EnsembleLoss",
                kwargs={"crps_weight": 1.0},
            )
        if self.n_ensemble == -1:
            if self.loss.type == "EnsembleLoss":
                self.n_ensemble = 2
            else:
                self.n_ensemble = 1

    @property
    def n_ic_timesteps(self) -> int:
        return self.step.n_ic_timesteps

    def get_evaluation_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return DataRequirements(
            names=self.all_names,
            n_timesteps=self._window_steps_required(n_forward_steps),
        )

    def get_prognostic_state_data_requirements(self) -> PrognosticStateDataRequirements:
        return PrognosticStateDataRequirements(
            names=self.prognostic_names,
            n_timesteps=self.n_ic_timesteps,
        )

    @property
    def input_only_names(self) -> list[str]:
        return list(set(self.input_names) - set(self.output_names))

    def get_forcing_window_data_requirements(
        self, n_forward_steps: int
    ) -> DataRequirements:
        return DataRequirements(
            names=list(
                set(self.input_only_names).union(self.step.next_step_input_names)
            ),
            n_timesteps=self._window_steps_required(n_forward_steps),
        )

    def _window_steps_required(self, n_forward_steps: int) -> int:
        return n_forward_steps + self.n_ic_timesteps

    def as_loaded_dict(self):
        self.step.load()
        return dataclasses.asdict(self)

    def get_stepper(
        self,
        dataset_info: DatasetInfo,
        apply_parameter_init: bool = True,
        training_history: TrainingHistory | None = None,
    ):
        """
        Args:
            dataset_info: Information about the training dataset.
            apply_parameter_init: Whether to apply parameter initialization.
            training_history: History of the stepper's training jobs.
        """
        logging.info("Initializing stepper from provided config")
        if apply_parameter_init:
            parameter_initializer = self.get_parameter_initializer()
        else:
            parameter_initializer = ParameterInitializer()
        step = self.step.get_step(
            dataset_info, init_weights=parameter_initializer.freeze_weights
        )
        derive_func = dataset_info.vertical_coordinate.build_derive_function(
            dataset_info.timestep
        )
        if self.input_masking is None:
            input_masking = NullMasking()
        else:
            input_masking = self.input_masking.build(
                mask=dataset_info.mask_provider,
                means=step.normalizer.means,
            )
        try:
            output_process_func = dataset_info.mask_provider.build_output_masker()
        except MissingDatasetInfo:
            output_process_func = NullPostProcessFn()
        return Stepper(
            config=self,
            step=step,
            dataset_info=dataset_info,
            input_process_func=input_masking,
            output_process_func=output_process_func,
            derive_func=derive_func,
            parameter_initializer=parameter_initializer,
            training_history=training_history,
        )

    @classmethod
    def from_stepper_state(cls, state) -> "StepperConfig":
        """
        Initialize a StepperConfig from a stepper state.

        This is required for backwards compatibility with older steppers,
        whose configuration did not provide normalization constants, but rather
        pointed to files on disk. Newer stepper configurations load these
        constants into the configuration before checkpoints are saved.

        Args:
            state: The state of the stepper.

        Returns:
            The stepper config.
        """
        try:
            legacy_config = SingleModuleStepperConfig.from_state(state["config"])
            normalizer = StandardNormalizer.from_state(
                state.get("normalizer", state.get("normalization"))
            )
            if normalizer is None:
                raise KeyError(
                    "No normalization found in state, available keys: "
                    + ", ".join(state.keys())
                )
            loss_normalizer_config = state.get(
                "loss_normalizer", state.get("loss_normalization")
            )
            if loss_normalizer_config is None:
                loss_normalizer = normalizer
            else:
                loss_normalizer = StandardNormalizer.from_state(loss_normalizer_config)
            return legacy_config.to_stepper_config(
                normalizer=normalizer, loss_normalizer=loss_normalizer
            )
        except (dacite.exceptions.DaciteError, KeyError):
            state = cls.remove_deprecated_keys(state["config"])
            return dacite.from_dict(
                data_class=cls, data=state, config=dacite.Config(strict=True)
            )

    @property
    def loss_names(self):
        """Names of variables to include in loss."""
        return self.step.loss_names

    @property
    def input_names(self) -> list[str]:
        """Names of variables which are required as inputs."""
        return self.step.input_names

    @property
    def all_names(self) -> list[str]:
        """Names of all variables."""
        return list(set(self.input_names + self.output_names))

    @property
    def next_step_forcing_names(self) -> list[str]:
        """
        Names of variables which are given as inputs but taken from the output timestep.

        An example might be solar insolation taken during the output window period.
        """
        return self.step.get_next_step_forcing_names()

    @property
    def prognostic_names(self) -> list[str]:
        """Names of variables which both inputs and outputs."""
        return self.step.prognostic_names

    @property
    def output_names(self) -> list[str]:
        """Names of variables which are outputs only."""
        return self.step.output_names

    @classmethod
    def remove_deprecated_keys(cls, state: dict[str, Any]) -> dict[str, Any]:
        state_copy = state.copy()
        return state_copy

    def replace_ocean(self, ocean: OceanConfig | None):
        self.step.replace_ocean(ocean)

    def get_ocean(self) -> OceanConfig | None:
        return self.step.get_ocean()

    def replace_multi_call(
        self, multi_call: MultiCallConfig | None, state: dict[str, Any]
    ) -> dict[str, Any]:
        """Replace the multi-call configuration of self.step and ensure the
        associated state can be loaded as a multi-call step.

        A value of `None` for `multi_call` will remove the multi-call configuration.

        If the selected type supports it, the multi-call configuration will be
        updated in place. Otherwise, it will be wrapped in the multi_call step
        configuration with the given multi_call config or None.

        Note this updates self.step in place, but returns a new state dictionary.

        Args:
            multi_call: MultiCallConfig for the resulting self.step.
            state: state dictionary associated with the loaded step.

        Returns:
            The state dictionary updated to ensure consistency with that of a
            serialized multi-call step.
        """
        self.step, new_state = replace_multi_call(self.step, multi_call, state)
        return new_state

    def get_parameter_initializer(self) -> ParameterInitializer:
        """
        Get the parameter initializer for this stepper configuration.
        """
        return self.parameter_init.build(
            load_weights_and_history=_load_weights_and_history
        )


class Stepper(
    TrainStepperABC[
        PrognosticState,
        BatchData,
        BatchData,
        PairedData,
        TrainOutput,
    ]
):
    """
    Stepper class for selectable step configurations.
    """

    TIME_DIM = 1
    CHANNEL_DIM = -3

    def __init__(
        self,
        config: StepperConfig,
        step: StepABC,
        dataset_info: DatasetInfo,
        input_process_func: Callable[[TensorMapping], TensorDict],
        output_process_func: Callable[[TensorMapping], TensorDict],
        derive_func: Callable[[TensorMapping, TensorMapping], TensorDict],
        parameter_initializer: ParameterInitializer,
        training_history: TrainingHistory | None = None,
    ):
        """
        Args:
            config: The configuration.
            step: The step object.
            dataset_info: Information about dataset used for training.
            output_process_func: Function to post-process the output of the step
                function.
            derive_func: Function to compute derived variables.
            input_process_func: Optional function for processing inputs and next-step
                inputs before passing them to the step object, e.g., by masking
                specific regions.
            parameter_initializer: The parameter initializer to use for loading weights
                from an external source.
            training_history: History of the stepper's training jobs.
        """
        self._config = config
        self._step_obj = step
        self._dataset_info = dataset_info
        self._derive_func = derive_func
        self._output_process_func = output_process_func
        self._input_process_func = input_process_func
        self._no_optimization = NullOptimization()
        self._parameter_initializer = parameter_initializer

        def get_loss_obj():
            loss_normalizer = step.get_loss_normalizer()
            if config.loss is None:
                raise ValueError("Loss is not configured")
            return config.loss.build(
                dataset_info.gridded_operations,
                out_names=config.loss_names,
                channel_dim=self.CHANNEL_DIM,
                normalizer=loss_normalizer,
            )

        self._loss_normalizer: StandardNormalizer | None = None

        self._get_loss_obj = get_loss_obj
        self._loss_obj: WeightedMappingLoss | None = None

        self._parameter_initializer.apply_weights(
            step.modules,
        )

        self._l2_sp_tuning_regularizer = (
            self._parameter_initializer.get_l2_sp_tuning_regularizer(
                step.modules,
            )
        )

        self._training_history = (
            training_history if training_history is not None else TrainingHistory()
        )
        self._append_training_history_from(
            base_training_history=self._parameter_initializer.training_history
        )

        _1: PredictFunction[  # for type checking
            PrognosticState,
            BatchData,
            BatchData,
        ] = self.predict

        _2: PredictFunction[  # for type checking
            PrognosticState,
            BatchData,
            PairedData,
        ] = self.predict_paired

        self._dataset_info = dataset_info

    @property
    def _loaded_loss_normalizer(self) -> StandardNormalizer:
        if self._loss_normalizer is None:
            loss_normalizer = self._step_obj.get_loss_normalizer()
            self._loss_normalizer = loss_normalizer
        return self._loss_normalizer

    @property
    def loss_obj(self) -> WeightedMappingLoss:
        if self._loss_obj is None:
            self._loss_obj = self._get_loss_obj()
        return self._loss_obj

    @property
    def config(self) -> StepperConfig:
        return self._config

    @property
    def derive_func(self) -> Callable[[TensorMapping, TensorMapping], TensorDict]:
        return self._derive_func

    @property
    def surface_temperature_name(self) -> str | None:
        return self._step_obj.surface_temperature_name

    @property
    def ocean_fraction_name(self) -> str | None:
        return self._step_obj.ocean_fraction_name

    @property
    def training_dataset_info(self) -> DatasetInfo:
        return self._dataset_info

    @property
    def training_variable_metadata(self) -> Mapping[str, VariableMetadata]:
        return self._dataset_info.variable_metadata

    @property
    def training_history(self) -> TrainingHistory:
        return self._training_history

    def _append_training_history_from(
        self, base_training_history: TrainingHistory | None
    ):
        """
        When the stepper receives weights from a base stepper via parameter
        initialization, this helper is used to extend its training history to include
        the training history of the base stepper.

        Args:
            base_training_history: The training history from a base stepper to append.
        """
        if base_training_history is not None:
            self._training_history.extend(base_training_history)

    @property
    def effective_loss_scaling(self) -> TensorDict:
        """
        Effective loss scalings used to normalize outputs before computing loss.
        y_loss_normalized_i = (y_i - y_mean_i) / loss_scaling_i
        where loss_scaling_i = loss_normalizer_std_i / weight_i.
        """
        return self.loss_obj.effective_loss_scaling

    def replace_multi_call(self, multi_call: MultiCallConfig | None):
        """
        Replace the MultiCall object with a new one. Note this is only
        meant to be used at inference time and may result in the loss
        function being unusable.

        Args:
            multi_call: The new multi_call configuration or None.
        """
        state = self._step_obj.get_state()
        new_state = self._config.replace_multi_call(multi_call, state)
        new_stepper: Stepper = self._config.get_stepper(
            dataset_info=self._dataset_info, apply_parameter_init=False
        )
        new_stepper._step_obj.load_state(new_state)
        self._step_obj = new_stepper._step_obj

    def replace_ocean(self, ocean: OceanConfig | None):
        """
        Replace the ocean model with a new one.

        Args:
            ocean: The new ocean model configuration or None.
        """
        self._config.replace_ocean(ocean)
        new_stepper: Stepper = self._config.get_stepper(
            dataset_info=self._dataset_info,
            apply_parameter_init=False,
        )
        new_stepper._step_obj.load_state(self._step_obj.get_state())
        self._step_obj = new_stepper._step_obj

    def get_base_weights(self) -> Weights | None:
        """
        Get the base weights of the stepper.

        Returns:
            A list of weight dictionaries for each module in the stepper.
        """
        return self._parameter_initializer.base_weights

    @property
    def prognostic_names(self) -> list[str]:
        return self._step_obj.prognostic_names

    @property
    def out_names(self) -> list[str]:
        return self._step_obj.output_names

    @property
    def loss_names(self) -> list[str]:
        return self._step_obj.loss_names

    @property
    def n_ic_timesteps(self) -> int:
        return self._step_obj.n_ic_timesteps

    @property
    def modules(self) -> nn.ModuleList:
        """
        Returns:
            A list of modules being trained.
        """
        return self._step_obj.modules

    @property
    def normalizer(self) -> StandardNormalizer:
        return self._step_obj.normalizer

    def step(
        self,
        input: TensorMapping,
        next_step_input_data: TensorMapping,
        wrapper: Callable[[nn.Module], nn.Module] = lambda x: x,
    ) -> TensorDict:
        """
        Step the model forward one timestep given input data.

        Args:
            input: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon] containing denormalized data from the
                initial timestep.
            next_step_input_data: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon] containing denormalized data from
                the output timestep.
            wrapper: Wrapper to apply over each nn.Module before calling.

        Returns:
            The denormalized output data at the next time step.
        """
        input = self._input_process_func(input)
        next_step_input_data = self._input_process_func(next_step_input_data)
        output = self._step_obj.step(input, next_step_input_data, wrapper=wrapper)
        return self._output_process_func(output)

    def get_prediction_generator(
        self,
        initial_condition: PrognosticState,
        forcing_data: BatchData,
        n_forward_steps: int,
        optimizer: OptimizationABC,
    ) -> Generator[TensorDict, None, None]:
        """
        Predict multiple steps forward given initial condition and forcing data.

        Uses low-level inputs and does not compute derived variables, to separate
        concerns from the public `predict` method.

        Args:
            initial_condition: The initial condition, containing tensors of shape
                [n_batch, self.n_ic_timesteps, <horizontal_dims>].
            forcing_data: The forcing data, containing tensors of shape
                [n_batch, n_forward_steps + self.n_ic_timesteps, <horizontal_dims>].
            n_forward_steps: The number of forward steps to predict, corresponding
                to the data shapes of forcing_data.
            optimizer: The optimizer to use for updating the module.

        Returns:
            Generator yielding the output data at each timestep.
        """
        ic_dict = initial_condition.as_batch_data().data
        forcing_dict = forcing_data.data
        return self._predict_generator(
            ic_dict, forcing_dict, n_forward_steps, optimizer
        )

    @property
    def _input_only_names(self) -> list[str]:
        return list(
            set(self._step_obj.input_names).difference(set(self._step_obj.output_names))
        )

    def _predict_generator(
        self,
        ic_dict: TensorMapping,
        forcing_dict: TensorMapping,
        n_forward_steps: int,
        optimizer: OptimizationABC,
    ) -> Generator[TensorDict, None, None]:
        state = {k: ic_dict[k].squeeze(self.TIME_DIM) for k in ic_dict}
        for step in range(n_forward_steps):
            input_forcing = {
                k: (
                    forcing_dict[k][:, step]
                    if k not in self._step_obj.next_step_forcing_names
                    else forcing_dict[k][:, step + 1]
                )
                for k in self._input_only_names
            }
            next_step_input_dict = {
                k: forcing_dict[k][:, step + 1]
                for k in self._step_obj.next_step_input_names
            }
            input_data = {**state, **input_forcing}

            def checkpoint(module):
                return optimizer.checkpoint(module, step=step)

            state = self.step(
                input_data,
                next_step_input_dict,
                wrapper=checkpoint,
            )
            yield state
            state = optimizer.detach_if_using_gradient_accumulation(state)

    def predict(
        self,
        initial_condition: PrognosticState,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> tuple[BatchData, PrognosticState]:
        """
        Predict multiple steps forward given initial condition and reference data.

        Args:
            initial_condition: Prognostic state data with tensors of shape
                [n_batch, self.n_ic_timesteps, <horizontal_dims>]. This data is assumed
                to contain all prognostic variables and be denormalized.
            forcing: Contains tensors of shape
                [n_batch, self.n_ic_timesteps + n_forward_steps, n_lat, n_lon]. This
                contains the forcing and ocean data for the initial condition and all
                subsequent timesteps.
            compute_derived_variables: Whether to compute derived variables for the
                prediction.

        Returns:
            A batch data containing the prediction and the prediction's final state
            which can be used as a new initial condition.
        """
        timer = GlobalTimer.get_instance()
        forcing_names = set(self._input_only_names).union(
            self._step_obj.next_step_input_names
        )
        with timer.context("forward_prediction"):
            forcing_data = forcing.subset_names(forcing_names)
            if initial_condition.as_batch_data().n_timesteps != self.n_ic_timesteps:
                raise ValueError(
                    f"Initial condition must have {self.n_ic_timesteps} timesteps, got "
                    f"{initial_condition.as_batch_data().n_timesteps}."
                )
            n_forward_steps = forcing_data.n_timesteps - self.n_ic_timesteps
            output_list = list(
                self.get_prediction_generator(
                    initial_condition,
                    forcing_data,
                    n_forward_steps,
                    NullOptimization(),
                )
            )
        data = process_prediction_generator_list(
            output_list,
            time=forcing_data.time[:, self.n_ic_timesteps :],
            horizontal_dims=forcing_data.horizontal_dims,
        )
        if compute_derived_variables:
            with timer.context("compute_derived_variables"):
                data = (
                    data.prepend(initial_condition)
                    .compute_derived_variables(
                        derive_func=self.derive_func,
                        forcing_data=forcing_data,
                    )
                    .remove_initial_condition(self.n_ic_timesteps)
                )
        prognostic_state = data.get_end(self.prognostic_names, self.n_ic_timesteps)
        data = BatchData.new_on_device(
            data=data.data,
            time=data.time,
            horizontal_dims=data.horizontal_dims,
        )
        return data, prognostic_state

    def predict_paired(
        self,
        initial_condition: PrognosticState,
        forcing: BatchData,
        compute_derived_variables: bool = False,
    ) -> tuple[PairedData, PrognosticState]:
        """
        Predict multiple steps forward given initial condition and reference data.

        Args:
            initial_condition: Prognostic state data with tensors of shape
                [n_batch, self.n_ic_timesteps, <horizontal_dims>]. This data is assumed
                to contain all prognostic variables and be denormalized.
            forcing: Contains tensors of shape
                [n_batch, self.n_ic_timesteps + n_forward_steps, n_lat, n_lon]. This
                contains the forcing and ocean data for the initial condition and all
                subsequent timesteps.
            compute_derived_variables: Whether to compute derived variables for the
                prediction.

        Returns:
            A tuple of 1) a paired data object, containing the prediction paired with
            all target/forcing data at the same timesteps, and 2) the prediction's
            final state, which can be used as a new initial condition.
        """
        prediction, new_initial_condition = self.predict(
            initial_condition, forcing, compute_derived_variables
        )
        forward_data = self.get_forward_data(
            forcing, compute_derived_variables=compute_derived_variables
        )
        return (
            PairedData.from_batch_data(
                prediction=prediction,
                reference=BatchData.new_on_device(
                    data=forward_data.data,
                    time=forward_data.time,
                    horizontal_dims=forward_data.horizontal_dims,
                ),
            ),
            new_initial_condition,
        )

    def get_forward_data(
        self, data: BatchData, compute_derived_variables: bool = False
    ) -> BatchData:
        if compute_derived_variables:
            timer = GlobalTimer.get_instance()
            with timer.context("compute_derived_variables"):
                data = data.compute_derived_variables(
                    derive_func=self.derive_func,
                    forcing_data=data,
                )
        return data.remove_initial_condition(self.n_ic_timesteps)

    def _get_regularizer_loss(self) -> torch.Tensor:
        return self._l2_sp_tuning_regularizer() + self._step_obj.get_regularizer_loss()

    def train_on_batch(
        self,
        data: BatchData,
        optimization: OptimizationABC,
        compute_derived_variables: bool = False,
    ) -> TrainOutput:
        """
        Train the model on a batch of data with one or more forward steps.

        If gradient accumulation is used by the optimization, the computational graph is
        detached between steps to reduce memory consumption. This means the model learns
        how to deal with inputs on step N but does not try to improve the behavior at
        step N by modifying the behavior for step N-1.

        Args:
            data: The batch data where each tensor in data.data has shape
                [n_sample, n_forward_steps + self.n_ic_timesteps, <horizontal_dims>].
            optimization: The optimization class to use for updating the module.
                Use `NullOptimization` to disable training.
            compute_derived_variables: Whether to compute derived variables for the
                prediction and target data.

        Returns:
            The loss metrics, the generated data, the normalized generated data,
                and the normalized batch data.
        """
        metrics: dict[str, float] = {}
        input_data = data.get_start(self.prognostic_names, self.n_ic_timesteps)
        target_data = self.get_forward_data(data, compute_derived_variables=False)

        optimization.set_mode(self._step_obj.modules)
        output_list = self._accumulate_loss(
            input_data,
            data,
            target_data,
            optimization,
            metrics,
        )

        regularizer_loss = self._get_regularizer_loss()
        if torch.any(regularizer_loss > 0):
            optimization.accumulate_loss(regularizer_loss)
        metrics["loss"] = optimization.get_accumulated_loss().detach()
        optimization.step_weights()

        gen_data = process_ensemble_prediction_generator_list(output_list)

        stepped = TrainOutput(
            metrics=metrics,
            gen_data=gen_data,
            target_data=add_ensemble_dim(target_data.data),
            time=target_data.time,
            normalize=self.normalizer.normalize,
            derive_func=self.derive_func,
        )
        ic = data.get_start(
            set(data.data.keys()), self.n_ic_timesteps
        )  # full data and not just prognostic get prepended
        stepped = stepped.prepend_initial_condition(ic)
        if compute_derived_variables:
            stepped = stepped.compute_derived_variables()
        # apply post-processing and return
        return stepped

    def _accumulate_loss(
        self,
        input_data: PrognosticState,
        data: BatchData,
        target_data: BatchData,
        optimization: OptimizationABC,
        metrics: dict[str, float],
    ) -> list[EnsembleTensorDict]:
        input_data = data.get_start(self.prognostic_names, self.n_ic_timesteps)
        # output from self.predict_paired does not include initial condition
        n_forward_steps = data.time.shape[1] - self.n_ic_timesteps
        n_ensemble = self._config.n_ensemble
        input_ensemble_data: TensorMapping = repeat_interleave_batch_dim(
            input_data.as_batch_data().data, repeats=n_ensemble
        )
        forcing_ensemble_data: TensorMapping = repeat_interleave_batch_dim(
            data.data, repeats=n_ensemble
        )
        output_generator = self._predict_generator(
            input_ensemble_data,
            forcing_ensemble_data,
            n_forward_steps,
            optimization,
        )
        output_list: list[EnsembleTensorDict] = []
        for step, gen_step in enumerate(output_generator):
            gen_step = unfold_ensemble_dim(gen_step, n_ensemble=n_ensemble)
            output_list.append(gen_step)
            # Note: here we examine the loss for a single timestep,
            # not a single model call (which may contain multiple timesteps).
            target_step = add_ensemble_dim(
                {k: v.select(self.TIME_DIM, step) for k, v in target_data.data.items()}
            )
            step_loss = self.loss_obj(gen_step, target_step)
            metrics[f"loss_step_{step}"] = step_loss.detach()
            optimization.accumulate_loss(step_loss)
        return output_list

    def update_training_history(self, training_job: TrainingJob) -> None:
        """
        Update the stepper's history of training jobs.

        Args:
            training_job: The training job to add to the history.
        """
        self._training_history.append(training_job)

    def get_state(self):
        """
        Returns:
            The state of the stepper.
        """
        return {
            "config": self._config.as_loaded_dict(),
            "dataset_info": self._dataset_info.to_state(),
            "step": self._step_obj.get_state(),
            "training_history": self._training_history.get_state(),
        }

    def load_state(self, state: dict[str, Any]) -> None:
        """
        Load the state of the stepper.

        Args:
            state: The state to load.
        """
        self._step_obj.load_state(state["step"])

    @classmethod
    def from_state(cls, state) -> "Stepper":
        """
        Load the state of the stepper.

        Args:
            state: The state to load.

        Returns:
            The stepper.
        """
        try:
            legacy_config = SingleModuleStepperConfig.from_state(state["config"])
            dataset_state = {}
            dataset_state["timestep"] = state.get(
                "encoded_timestep", DEFAULT_ENCODED_TIMESTEP
            )
            if "sigma_coordinates" in state:
                # for backwards compatibility with old checkpoints
                dataset_state["vertical_coordinate"] = state["sigma_coordinates"]
            else:
                dataset_state["vertical_coordinate"] = state["vertical_coordinate"]

            if "area" in state:
                # backwards-compatibility, these older checkpoints are always lat-lon
                dataset_state["gridded_operations"] = {
                    "type": "LatLonOperations",
                    "state": {"area_weights": state["area"]},
                }
            else:
                dataset_state["gridded_operations"] = state["gridded_operations"]

            if "img_shape" in state:
                dataset_state["img_shape"] = state["img_shape"]

            normalizer = StandardNormalizer.from_state(
                state.get("normalizer", state.get("normalization"))
            )
            if normalizer is None:
                raise ValueError(
                    f"No normalizer state found, keys include {state.keys()}"
                )
            loss_normalizer = StandardNormalizer.from_state(
                state.get("loss_normalizer", state.get("loss_normalization"))
            )
            if loss_normalizer is None:
                loss_normalizer = normalizer
            config = legacy_config.to_stepper_config(
                normalizer=normalizer, loss_normalizer=loss_normalizer
            )
            dataset_info = DatasetInfo.from_state(dataset_state)
            state["step"] = {
                # SingleModuleStep inside MultiCallStep
                "wrapped_step": {"module": state["module"]}
            }
        except dacite.exceptions.DaciteError:
            config = StepperConfig.from_stepper_state(state)
            dataset_info = DatasetInfo.from_state(state["dataset_info"])
        training_history = TrainingHistory.from_state(state.get("training_history", []))
        stepper = config.get_stepper(
            dataset_info=dataset_info,
            training_history=training_history,
            # don't need to initialize weights, we're about to load_state
            apply_parameter_init=False,
        )
        stepper.load_state(state)
        return stepper


def get_serialized_stepper_vertical_coordinate(
    state: dict[str, Any],
) -> VerticalCoordinate:
    if "vertical_coordinate" in state:
        return dacite.from_dict(
            data_class=SerializableVerticalCoordinate,
            data={"vertical_coordinate": state["vertical_coordinate"]},
            config=dacite.Config(strict=True),
        ).vertical_coordinate
    elif "sigma_coordinates" in state:
        return dacite.from_dict(
            data_class=SerializableVerticalCoordinate,
            data={"vertical_coordinate": state["sigma_coordinates"]},
            config=dacite.Config(strict=True),
        ).vertical_coordinate
    else:
        dataset_info = DatasetInfo.from_state(state["dataset_info"])
        return dataset_info.vertical_coordinate


@dataclasses.dataclass
class StepperOverrideConfig:
    """
    Configuration for overriding stepper configuration options.

    The default value for each parameter is ``"keep"``, which denotes that the
    serialized stepper's configuration will not be modified when loaded. Passing
    other values will override the configuration of the loaded stepper.

    Parameters:
        ocean: Ocean configuration to override that used in producing a serialized
            stepper.
        multi_call: MultiCall configuration to override that used in producing a
            serialized stepper.
    """

    ocean: Literal["keep"] | OceanConfig | None = "keep"
    multi_call: Literal["keep"] | MultiCallConfig | None = "keep"


def load_stepper_config(
    checkpoint_path: str | pathlib.Path,
    override_config: StepperOverrideConfig | None = None,
) -> StepperConfig:
    """Load a stepper configuration, optionally overriding certain aspects.

    Args:
        checkpoint_path: The path to the serialized checkpoint.
        override_config: Configuration options to override (optional).

    Returns:
        The configuration of the stepper serialized in the checkpoint, with
        appropriate options overridden.
    """
    stepper = load_stepper(checkpoint_path, override_config)
    return stepper._config


def load_stepper(
    checkpoint_path: str | pathlib.Path,
    override_config: StepperOverrideConfig | None = None,
) -> Stepper:
    """Load a stepper, optionally overriding certain aspects.

    Args:
        checkpoint_path: The path to the serialized checkpoint.
        override_config: Configuration options to override (optional).

    Returns:
        The stepper serialized in the checkpoint, with appropriate options
        overridden.
    """
    if override_config is None:
        override_config = StepperOverrideConfig()

    checkpoint = torch.load(
        checkpoint_path, map_location=get_device(), weights_only=False
    )
    stepper = Stepper.from_state(checkpoint["stepper"])

    if override_config.ocean != "keep":
        logging.info(
            "Overriding training ocean configuration with a new ocean configuration."
        )
        stepper.replace_ocean(override_config.ocean)

    if override_config.multi_call != "keep":
        logging.info(
            "Overriding training multi_call configuration with a new "
            "multi_call configuration."
        )
        stepper.replace_multi_call(override_config.multi_call)
    return stepper



import dataclasses

import pytest
import torch

from fme.core.ensemble import get_crps


@dataclasses.dataclass
class CRPSExperiment:
    name: str
    truth_amount: float
    random_amount: float


@pytest.mark.parametrize("alpha", [1.0, 0.95, 0.1])
def test_crps(alpha: float):
    """
    Test that get_crps is a proper scoring rule.

    Scoring rules that are proper are proven to have the lowest
    expected score if the predicted distribution equals the
    underlying distribution of the target variable.
    """
    torch.manual_seed(0)
    nx = 1
    ny = 1
    n_batch = 10000
    n_sample = 2
    truth_amount = 0.8
    random_amount = 0.5
    experiments = [
        CRPSExperiment("perfect", truth_amount, random_amount),
        CRPSExperiment("extra_variance", truth_amount, random_amount * 1.1),
        CRPSExperiment("less_variance", truth_amount, random_amount * 0.9),
        CRPSExperiment("deterministic", truth_amount, random_amount * 1e-5),
    ]
    x_predictable = torch.rand(n_batch, 1, nx, ny)
    x = truth_amount * x_predictable + random_amount * torch.rand(n_batch, 1, nx, ny)
    crps_values = {}
    for experiment in experiments:
        x_sample = (
            experiment.truth_amount * x_predictable
            + experiment.random_amount * torch.rand(n_batch, n_sample, nx, ny)
        )
        crps_values[experiment.name] = get_crps(
            gen=x_sample, target=x, alpha=alpha
        ).mean()
    assert crps_values["perfect"] < crps_values["extra_variance"]
    assert crps_values["perfect"] < crps_values["less_variance"]
    assert crps_values["extra_variance"] < crps_values["deterministic"]
    assert crps_values["less_variance"] < crps_values["deterministic"]



from typing import Any

import pytest
import torch

from fme.core.gridded_ops import GriddedOperations, HEALPixOperations, LatLonOperations
from fme.core.typing_ import TensorMapping


@pytest.mark.parametrize(
    "state, expected_class",
    [
        (
            {
                "type": "LatLonOperations",
                "state": {"area_weights": torch.tensor([[1.0, 1.0], [1.0, 1.0]])},
            },
            LatLonOperations,
        ),
        (
            {
                "type": "HEALPixOperations",
                "state": {},
            },
            HEALPixOperations,
        ),
    ],
)
def test_gridded_operations_from_state(
    state: dict[str, Any],
    expected_class: type[GriddedOperations],
):
    ops = GriddedOperations.from_state(state)
    assert isinstance(ops, expected_class)

    recovered_state = ops.to_state()
    assert recovered_state == state

    with pytest.raises(RuntimeError):
        expected_class.from_state(state["state"])


def test_latlon_area_weighted_sum_dict():
    area_weights = torch.ones(4, 5)
    ops = LatLonOperations(area_weights=area_weights)
    data_dict: TensorMapping = {
        "var1": torch.randn(2, 4, 5),
        "var2": torch.randn(2, 4, 5),
    }
    result = ops.area_weighted_sum_dict(data_dict)
    assert isinstance(result, dict)
    assert set(result.keys()) == {"var1", "var2"}
    for var_name in data_dict:
        assert result[var_name].shape == (2,)
        # all weights are 1, so it's just sum
        torch.testing.assert_close(
            result[var_name], data_dict[var_name].sum(dim=(-2, -1))
        )


def test_latlon_zonal_mean():
    area_weights = torch.ones(4, 5)
    ops = LatLonOperations(area_weights=area_weights)
    n_sample, n_time, n_lat, n_lon = 2, 4, 5, 6
    data = torch.randn(n_sample, n_time, n_lat, n_lon)
    result = ops.zonal_mean(data)
    assert result.shape == (n_sample, n_time, n_lat)
    torch.testing.assert_close(result, data.mean(dim=-1))


def test_latlon_area_weighted_mean_dict():
    area_weights = torch.ones(4, 5)
    ops = LatLonOperations(area_weights=area_weights)
    data_dict: TensorMapping = {
        "var1": torch.randn(2, 4, 5),
        "var2": torch.randn(2, 4, 5),
    }
    result = ops.area_weighted_mean_dict(data_dict)
    assert isinstance(result, dict)
    assert set(result.keys()) == {"var1", "var2"}
    for var_name in data_dict:
        assert result[var_name].shape == (2,)
        # all weights are 1, so it's just mean
        torch.testing.assert_close(
            result[var_name], data_dict[var_name].mean(dim=(-2, -1))
        )


def test_latlon_area_weighted_rmse_dict():
    area_weights = torch.rand(4).unsqueeze(-1).broadcast_to(4, 5)
    ops = LatLonOperations(area_weights=area_weights)
    truth_dict: TensorMapping = {
        "var1": torch.randn(2, 4, 5),
        "var2": torch.randn(2, 4, 5),
    }
    predicted_dict: TensorMapping = {
        "var1": torch.randn(2, 4, 5),
        "var2": torch.randn(2, 4, 5),
    }
    result = ops.area_weighted_rmse_dict(truth_dict, predicted_dict)
    assert isinstance(result, dict)
    assert set(result.keys()) == {"var1", "var2"}

    for var_name in truth_dict:
        assert result[var_name].shape == (2,)
        diff_sq = (predicted_dict[var_name] - truth_dict[var_name]) ** 2
        weighted_diff_sq = diff_sq * area_weights
        mean_weighted_diff_sq = weighted_diff_sq.sum(dim=(-2, -1)) / area_weights.sum()
        expected_rmse_var1 = torch.sqrt(mean_weighted_diff_sq)
        torch.testing.assert_close(result[var_name], expected_rmse_var1)


def test_latlon_area_weighted_std_dict_input():
    area_weights = torch.rand(4).unsqueeze(-1).broadcast_to(4, 5)
    ops = LatLonOperations(area_weights=area_weights)
    data_dict: TensorMapping = {
        "var1": torch.randn(2, 4, 5),
        "var2": torch.randn(2, 4, 5),
    }
    result = ops.area_weighted_std_dict(data_dict)
    assert isinstance(result, dict)
    assert set(result.keys()) == {"var1", "var2"}

    for var_name in data_dict:
        assert result[var_name].shape == (2,)
        mean_var1 = ops.area_weighted_mean(data_dict[var_name], keepdim=True)
        variance = ops.area_weighted_mean((data_dict[var_name] - mean_var1) ** 2)
        expected_std_var1 = variance.sqrt()
        torch.testing.assert_close(result[var_name], expected_std_var1)



import copy
import logging
from collections.abc import Mapping, Sequence
from pathlib import Path

import torch
import xarray as xr

from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.histogram import DynamicHistogram


class _HistogramAggregator:
    def __init__(self, n_times: int, names: Sequence[str] | None = None):
        self._histograms: Mapping[str, DynamicHistogram] | None = None
        self._n_times = n_times
        self._names = names

    def record_batch(
        self,
        data: dict[str, torch.Tensor],
        i_time_start: int,
    ):
        if self._histograms is None:
            if self._names is not None:
                save_names_not_in_data = set(self._names) - set(data.keys())
                if len(save_names_not_in_data) > 0:
                    logging.warning(
                        f"Variables {save_names_not_in_data} in save_names are not "
                        "in histogram data."
                    )
                _names = [k for k in data if k in self._names]
            else:
                _names = list(data)
            self._histograms = {
                var_name: DynamicHistogram(n_times=self._n_times) for var_name in _names
            }

        for var_name, histogram in self._histograms.items():
            # go from [n_samples, n_timesteps, n_lat, n_lon] to
            #     [n_timesteps, n_samples, n_lat, n_lon]
            # and then reshape to [n_timesteps, n_hist_samples]
            n_times = data[var_name].shape[1]
            reshaped_data = data[var_name].transpose(1, 0).reshape(n_times, -1)
            histogram.add(reshaped_data, i_time_start=i_time_start)

    def get_dataset(self) -> xr.Dataset:
        if self._histograms is None:
            raise RuntimeError("No data has been recorded.")
        return self._get_single_dataset(self._histograms)

    @staticmethod
    def _get_single_dataset(histograms: Mapping[str, DynamicHistogram]) -> xr.Dataset:
        data = {}
        for var_name, histogram in histograms.items():
            data[var_name] = xr.DataArray(
                histogram.counts,
                dims=("time", "bin"),
            )
            data[f"{var_name}_bin_edges"] = xr.DataArray(
                histogram.bin_edges,
                dims=("bin_edges",),
            )
        return xr.Dataset(data)


class PairedHistogramDataWriter:
    """
    Wrapper over HistogramDataWriter to write both target and prediction data.
    Gives the same interface as HistogramDataWriter.
    """

    def __init__(
        self,
        path: str,
        n_timesteps: int,
        variable_metadata: Mapping[str, VariableMetadata],
        save_names: Sequence[str] | None,
        dataset_metadata: DatasetMetadata,
    ):
        self._target_writer = HistogramDataWriter(
            path=path,
            n_timesteps=n_timesteps,
            filename="histograms_target.nc",
            variable_metadata=variable_metadata,
            save_names=save_names,
            dataset_metadata=dataset_metadata,
        )
        self._prediction_writer = HistogramDataWriter(
            path=path,
            n_timesteps=n_timesteps,
            filename="histograms_prediction.nc",
            variable_metadata=variable_metadata,
            save_names=save_names,
            dataset_metadata=dataset_metadata,
        )

    def append_batch(
        self,
        target: dict[str, torch.Tensor],
        prediction: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        self._target_writer.append_batch(
            data=target,
            start_timestep=start_timestep,
            batch_time=batch_time,
        )
        self._prediction_writer.append_batch(
            data=prediction,
            start_timestep=start_timestep,
            batch_time=batch_time,
        )

    def flush(self):
        self._target_writer.flush()
        self._prediction_writer.flush()


class HistogramDataWriter:
    """
    Write [time, bin] histogram data for each variable to a netCDF file.
    """

    def __init__(
        self,
        path: str,
        n_timesteps: int,
        filename: str,
        variable_metadata: Mapping[str, VariableMetadata],
        save_names: Sequence[str] | None,
        dataset_metadata: DatasetMetadata,
    ):
        """
        Args:
            path: The directory within which to write the file.
            n_timesteps: Number of timesteps to write to the file.
            filename: Name of the file to write.
            variable_metadata: Metadata for each variable to be written to the file.
            save_names: Names of variables to save. If None, all variables are saved.
            dataset_metadata: Metadata for the dataset.
        """
        self.path = path
        self._metrics_filename = str(Path(path) / filename)
        self.variable_metadata = variable_metadata
        self._histogram = _HistogramAggregator(n_times=n_timesteps, names=save_names)
        self._dataset_metadata = copy.copy(dataset_metadata)
        self._dataset_metadata.title = (
            f"ACE {filename.removesuffix('.nc').replace('_', ' ')} data file"
        )

    def append_batch(
        self,
        data: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        """
        Append a batch of data to the file.

        Args:
            data: The data to write.
            start_timestep: Timestep at which to start writing.
            batch_time: Time coordinate for each sample in the batch.
        """
        del batch_time
        self._histogram.record_batch(
            data=data,
            i_time_start=start_timestep,
        )

    def flush(self):
        """
        Flush the data to disk.
        """
        metric_dataset = self._histogram.get_dataset()
        for name in self.variable_metadata:
            try:
                metric_dataset[f"{name}_bin_edges"].attrs["units"] = (
                    self.variable_metadata[name].units
                )
            except KeyError:
                logging.info(
                    f"{name} in metadata but not in data written to "
                    f"{self._metrics_filename}."
                )
        for name in metric_dataset.data_vars:
            if not name.endswith("_bin_edges"):
                metric_dataset[f"{name}_bin_edges"].attrs["long_name"] = (
                    f"{name} bin edges"
                )
                metric_dataset[name].attrs["units"] = "count"
                metric_dataset[name].attrs["long_name"] = f"{name} histogram"
        metric_dataset.attrs.update(self._dataset_metadata.as_flat_str_dict())
        metric_dataset.to_netcdf(self._metrics_filename)



import os

import click
import numpy as np
import pandas as pd
import xarray as xr


@click.command()
@click.argument("input_zarr")
@click.argument("output_directory")
@click.option("--start-date", help="For subsetting, e.g. '2016-01-01'")
@click.option("--end-date", help="For subsetting, e.g. '2016-12-31'")
@click.option("--nc-format", default="NETCDF4", help="netCDF file format")
@click.option(
    "--prepend-nans",
    is_flag=True,
    help="Prepend NaNs to first timestep. Used for baseline "
    "which is missing initial condition.",
)
def main(input_zarr, output_directory, start_date, end_date, nc_format, prepend_nans):
    """Save data at INPUT_ZARR to monthly netcdf files in OUTPUT_DIRECTORY.
    It is assumed that OUTPUT_DIRECTORY does not exist."""
    os.makedirs(output_directory, exist_ok=True)
    ds = xr.open_zarr(input_zarr)
    if prepend_nans:
        # prepend NaNs to first timestep
        ds = prepend_nans_to_dataset(ds)
    ds = ds.sel(time=slice(start_date, end_date))
    monthly_ds = ds.resample(time="MS")
    for label, data in monthly_ds:
        if isinstance(label, np.datetime64):
            # np.datetime64 times do not have a strftime method,
            # so convert to pd.Timestamp
            label = pd.Timestamp(label)
        print(f"Processing month {label}")
        filename = os.path.join(output_directory, label.strftime("%Y%m%d%H") + ".nc")
        # use these options to enable opening data with netCDF4.MFDataset
        data.to_netcdf(filename, unlimited_dims=["time"], format=nc_format)


def prepend_nans_to_dataset(ds: xr.Dataset) -> xr.Dataset:
    """Prepend NaNs to time dimension of an xarray dataset."""
    time_dependent_vars = [v for v in ds if "time" in ds[v].dims]
    time_dependent_ds = ds[time_dependent_vars]
    prepend_step = xr.full_like(time_dependent_ds.isel(time=0), np.nan)
    delta_t = ds["time"].values[1] - ds["time"].values[0]
    prepend_step["time"] = ds["time"].values[0] - delta_t
    return xr.concat([prepend_step, ds], dim="time").transpose(*ds.dims)


def test_prepend_nans():
    ds = xr.tutorial.open_dataset("air_temperature")
    ds_prepended = prepend_nans_to_dataset(ds)
    assert ds_prepended.sizes["time"] == ds.sizes["time"] + 1
    assert np.isnan(ds_prepended.isel(time=0)["air"].values).all()


if __name__ == "__main__":
    main()



import pytest
import torch
import xarray as xr

from fme.core.dataset.time_coarsen import TimeCoarsenConfig, TimeCoarsenDataset
from fme.core.testing import date_range
from fme.core.typing_ import TensorDict


class BasicDataset(torch.utils.data.Dataset):
    def __init__(self, samples: list[tuple[TensorDict, xr.DataArray]]):
        if len(samples) == 0:
            raise ValueError("At least one sample is required")
        if len(set(len(sample[1]) for sample in samples)) != 1:
            raise ValueError("All samples must have the same number of timesteps")
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        return self.samples[idx]

    @property
    def sample_start_times(self) -> xr.CFTimeIndex:
        values = []
        for _, time in self.samples:
            values.append(time[0].values.item())
        return xr.CFTimeIndex(values)

    @property
    def sample_n_times(self) -> int:
        return self.samples[0][1].shape[0]


def get_sample(names: list[str], n_timesteps: int) -> tuple[TensorDict, xr.DataArray]:
    n_lat = 8
    n_lon = 16
    data = {name: torch.randn(n_timesteps, n_lat, n_lon) for name in names}
    times = date_range(n_timesteps)
    return (data, times)


@pytest.mark.parametrize("n_timesteps, factor", [(16, 2), (17, 2), (8, 1)])
def test_time_coarsen_dataset(n_timesteps: int, factor: int):
    config = TimeCoarsenConfig(
        factor=factor,
        snapshot_names=["snapshot"],
        window_names=["window"],
    )
    base_sample = get_sample(["snapshot", "window"], n_timesteps)
    base_data, base_time = base_sample
    base_dataset = BasicDataset([base_sample])
    dataset = TimeCoarsenDataset(base_dataset, config)
    assert len(dataset) == len(base_dataset)
    assert dataset.sample_n_times == base_dataset.sample_n_times // config.factor
    assert dataset.sample_start_times == base_dataset.sample_start_times
    for data, time in dataset:
        assert data["snapshot"].shape[0] == dataset.sample_n_times
        assert data["window"].shape[0] == dataset.sample_n_times
        assert time.shape[0] == dataset.sample_n_times
        assert time.values[0] == base_dataset.sample_start_times[0]
        assert time.values[0] == base_time.values[0]
        assert time.values[1] == base_time.values[config.factor]
        torch.testing.assert_close(
            data["snapshot"][0],
            base_data["snapshot"][config.factor - 1],
        )
        torch.testing.assert_close(
            data["snapshot"][1],
            base_data["snapshot"][config.factor * 2 - 1],
        )
        torch.testing.assert_close(
            data["window"][0],
            torch.mean(base_data["window"][: config.factor], dim=0),
        )
        torch.testing.assert_close(
            data["window"][1],
            torch.mean(base_data["window"][config.factor : 2 * config.factor], dim=0),
        )
